"""Define the Species class, used as a base for Gene and Metabolite."""


from copy import deepcopy
from typing import TYPE_CHECKING, FrozenSet, Optional

from ..core.object import Object


if TYPE_CHECKING:
    from .. import Model


class Species(Object):
    """Species is a base class in Cobrapy.

    Species is a class for holding information regarding
    a chemical Species

    Parameters
    ----------
    id : string
       An identifier for the chemical species
    name : string
       A human readable name.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self, id: Optional[str] = None, name: Optional[str] = None, **kwargs
    ) -> None:
        """Initialize a Species.

        Parameters
        ----------
        id : string, optional, default None
            An identifier for the chemical species
        name : string, optional, default None
            A human readable name.

        A species also contains a _model, reference to a cobra.model (initialized as
        None) and a self._reaction, a set of cobra.reactions (initialized as empty set).
        """
        super().__init__(id=id, name=name, **kwargs)
        self._model = None
        # references to reactions that operate on this species
        self._reaction = set()

    @property
    def reactions(self) -> FrozenSet:
        """Return a frozenset of reactions.

        Returns
        -------
        FrozenSet
            A frozenset that includes the reactions of the species.
        """
        return frozenset(self._reaction)

    def __getstate__(self) -> dict:
        """Return the state of the species.

        Remove the references to container reactions when serializing to
        avoid problems associated with recursion.

        Returns
        -------
        dict
            A dictionary describing the state, without the self._reaction to avoid
            recursion.
        """
        state = Object.__getstate__(self)
        state["_reaction"] = set()
        return state

    def copy(self) -> "Species":
        """Copy a species.

        When copying a reaction, it is necessary to deepcopy the
        components so the list references aren't carried over.

        Additionally, a copy of a reaction is no longer in a cobra.Model.

        This should be fixed with self.__deepcopy__ if possible

        Returns
        -------
        Species
            A copy of the species.
        """
        return deepcopy(self)

    @property
    def model(self) -> Optional["Model"]:
        """Return the model.

        Returns
        -------
        model
            Returns the cobra model that the species is associated with. None if there
            is no model associated with this species.
        """
        return self._model



"""Provide functions for reaction and gene deletions."""

from functools import partial
from itertools import product
from typing import TYPE_CHECKING, List, Optional, Set, Tuple, Union

import pandas as pd
from optlang.exceptions import SolverError

from ..core import Configuration, Gene, Model, Reaction
from ..util import ProcessPool
from ..util import solver as sutil
from .moma import add_moma
from .room import add_room


if TYPE_CHECKING:
    from cobra import Solution


configuration = Configuration()


def _get_growth(model: Model) -> Tuple[float, str]:
    """Return the growth from the `model`.

    Parameters
    ----------
    model : cobra.Model
        The model to obtain growth for.

    Returns
    -------
    float
        The obtained growth value. Returns nan if there is some error while
        optimizing.

    """
    try:
        if "moma_old_objective" in model.solver.variables:
            model.slim_optimize()
            growth = model.solver.variables.moma_old_objective.primal
        else:
            growth = model.slim_optimize()
    except SolverError:
        growth = float("nan")
    return growth, model.solver.status


def _reaction_deletion(
    model: Model, reaction_ids: List[str]
) -> Tuple[List[str], float, str]:
    """Perform reaction deletion.

    Parameters
    ----------
    model : cobra.Model
        The model to perform reaction deletion on.
    ids : list of str
        The reaction IDs to knock-out.

    Returns
    -------
    tuple of (list of str, float, str)
        A tuple containing reaction IDs knocked out, growth of the model and
        the solver status.

    """
    with model:
        for rxn_id in reaction_ids:
            model.reactions.get_by_id(rxn_id).knock_out()
        growth, status = _get_growth(model)
    return reaction_ids, growth, status


def _reaction_deletion_worker(ids: List[str]) -> Tuple[List[str], float, str]:
    """Perform reaction deletions on worker process.

    Parameters
    ----------
    ids : list of str
        The reaction IDs to knock-out.

    Returns
    -------
    tuple of (list of str, float, str)
        A tuple containing reaction IDs knocked out, growth of the model and
        the solver status.

    """
    global _model

    return _reaction_deletion(_model, ids)


def _gene_deletion(model: Model, gene_ids: List[str]) -> Tuple[List[str], float, str]:
    """Perform gene deletions.

    Parameters
    ----------
    model : cobra.Model
        The model to perform gene deletion on.
    ids : list of str
        The gene IDs to knock-out.

    Returns
    -------
    tuple of (list of str, float, str)
        A tuple containing gene IDs knocked out, growth of the model and
        the solver status.

    """
    with model:
        for gene_id in gene_ids:
            model.genes.get_by_id(gene_id).knock_out()
        growth, status = _get_growth(model)
    return gene_ids, growth, status


def _gene_deletion_worker(ids: List[str]) -> Tuple[List[str], float, str]:
    """Perform gene deletions on worker process.

    Parameters
    ----------
    ids : list of str
        The gene IDs to knock-out.

    Returns
    -------
    tuple of (list of str, float, str)
        A tuple containing gene IDs knocked out, growth of the model and
        the solver status.

    """
    global _model

    return _gene_deletion(_model, ids)


def _init_worker(model: Model) -> None:
    """Initialize worker process."""
    global _model

    _model = model


def _multi_deletion(
    model: Model,
    entity: str,
    element_lists: List[Union[Gene, Reaction]],
    method: str = "fba",
    solution: Optional["Solution"] = None,
    processes: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """Provide a common interface for single or multiple knockouts.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to perform deletions in.
    entity : {"gene", "reaction"}
        The entity to knockout.
    element_lists : list of cobra.Gene or cobra.Reaction
        List of cobra.Gene or cobra.Reaction to be deleted.
    method : {"fba", "moma", "linear moma", "room", "linear room"}, optional
        Method used to predict the growth rate (default "fba").
    solution : cobra.Solution, optional
        A previous solution to use as a reference for (linear) MOMA or ROOM
        (default None).
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not passed,
        will be set to `configuration.processes` (default None).
    **kwargs :
        Passed on to underlying simulation functions.

    Returns
    -------
    pandas.DataFrame
        A representation of all combinations of entity deletions. The
        columns are 'growth' and 'status', where

        index : tuple(str)
            The gene or reaction identifiers that were knocked out.
        growth : float
            The growth rate of the adjusted model.
        status : str
            The solution's status.

    """
    solver = sutil.interface_to_str(model.problem.__name__)
    if method == "moma" and solver not in sutil.qp_solvers:
        raise RuntimeError(
            f"Cannot use MOMA since '{solver}' is not QP-capable. "
            "Please choose a different solver or use FBA only."
        )

    if processes is None:
        processes = configuration.processes

    with model:
        if "moma" in method:
            add_moma(model, solution=solution, linear="linear" in method)
        elif "room" in method:
            add_room(model, solution=solution, linear="linear" in method, **kwargs)

        args = {frozenset(comb) for comb in product(*element_lists)}
        processes = min(processes, len(args))

        def extract_knockout_results(result_iter):
            result = pd.DataFrame(
                [
                    (
                        set(ids),
                        growth,
                        status,
                    )
                    for (ids, growth, status) in result_iter
                ],
                columns=["ids", "growth", "status"],
            )
            return result

        if processes > 1:
            worker = {
                "gene": _gene_deletion_worker,
                "reaction": _reaction_deletion_worker,
            }[entity]
            chunk_size = len(args) // processes

            with ProcessPool(
                processes, initializer=_init_worker, initargs=(model,)
            ) as pool:
                results = extract_knockout_results(
                    pool.imap_unordered(worker, args, chunksize=chunk_size)
                )
        else:
            worker = {"gene": _gene_deletion, "reaction": _reaction_deletion}[entity]
            results = extract_knockout_results(map(partial(worker, model), args))
        return results


def _entities_ids(entities: List[Union[str, Gene, Reaction]]) -> List[str]:
    """Return the IDs of the `entities`.

    Parameters
    ----------
    entities : list of str or cobra.Gene or cobra.Reaction
        The list of entities whose IDs need to be returned.

    Returns
    -------
    list of str
        The IDs of the `entities`.

    """
    try:
        return [e.id for e in entities]
    except AttributeError:
        return list(entities)


def _element_lists(
    entities: List[Union[str, Gene, Reaction]], *ids: List[str]
) -> List[str]:
    """Return the elements.

    Parameters
    ----------
    entities : list of str or cobra.Gene or cobra.Reaction
        The list of entities.
    *ids : list of str
        The list of IDs.

    Returns
    -------
    list of str
        The list of IDs.

    """
    lists = list(ids)
    if lists[0] is None:
        lists[0] = entities
    result = [_entities_ids(lists[0])]
    for _list in lists[1:]:
        if _list is None:
            result.append(result[-1])
        else:
            result.append(_entities_ids(_list))
    return result


def single_reaction_deletion(
    model: Model,
    reaction_list: Optional[List[Union[Reaction, str]]] = None,
    method: str = "fba",
    solution: Optional["Solution"] = None,
    processes: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """Knock out each reaction from `reaction_list`.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to perform deletions in.
    reaction_list : list of cobra.Reaction or str, optional
        The reactions be knocked out. If not passed, all the reactions from
        the model are used (default None).
    method: {"fba", "moma", "linear moma", "room", "linear room"}, optional
        Method used to predict the growth rate (default "fba").
    solution : cobra.Solution, optional
        A previous solution to use as a reference for (linear) MOMA or ROOM
        (default None).
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not passed,
        will be set to `configuration.processes` (default None).
    **kwargs :
        Keyword arguments are passed on to underlying simulation functions
        such as `add_room`.

    Returns
    -------
    pandas.DataFrame
        A representation of all single reaction deletions. The columns are
        'growth' and 'status', where

        index : tuple(str)
            The reaction identifier that was knocked out.
        growth : float
            The growth rate of the adjusted model.
        status : str
            The solution's status.

    """
    return _multi_deletion(
        model,
        "reaction",
        element_lists=_element_lists(model.reactions, reaction_list),
        method=method,
        solution=solution,
        processes=processes,
        **kwargs,
    )


def single_gene_deletion(
    model: Model,
    gene_list: Optional[List[Union[Gene, str]]] = None,
    method: str = "fba",
    solution: Optional["Solution"] = None,
    processes: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """Knock out each gene from `gene_list`.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to perform deletions in.
    gene_list : list of cobra.Gene or str, optional
        The gene objects to be deleted. If not passed, all the genes from the
        model are used (default None).
    method : {"fba", "moma", "linear moma", "room", "linear room"}, optional
        Method used to predict the growth rate (default "fba").
    solution : cobra.Solution, optional
        A previous solution to use as a reference for (linear) MOMA or ROOM
        (default None).
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not passed,
        will be set to `configuration.processes` (default None).
    **kwargs :
        Keyword arguments are passed on to underlying simulation functions
        such as `add_room`.

    Returns
    -------
    pandas.DataFrame
        A representation of all single gene deletions. The columns are
        'growth' and 'status', where

        index : tuple(str)
            The gene identifier that was knocked out.
        growth : float
            The growth rate of the adjusted model.
        status : str
            The solution's status.

    """
    return _multi_deletion(
        model,
        "gene",
        element_lists=_element_lists(model.genes, gene_list),
        method=method,
        solution=solution,
        processes=processes,
        **kwargs,
    )


def double_reaction_deletion(
    model: Model,
    reaction_list1: Optional[List[Union[Reaction, str]]] = None,
    reaction_list2: Optional[List[Union[Reaction, str]]] = None,
    method: str = "fba",
    solution: Optional["Solution"] = None,
    processes: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """Knock out each reaction pair from the combinations of two given lists.

    We say 'pair' here but the order order does not matter.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to perform deletions in.
    reaction_list1 : list of cobra.Reaction or str, optional
        The first reaction list to be deleted. If not passed,
        all the reactions from the model are used (default None).
    reaction_list2 : list of cobra.Reaction or str, optional
        The second reaction list to be deleted. If not passed,
        all the reactions from the model are used (default None).
    method: {"fba", "moma", "linear moma", "room", "linear room"}, optional
        Method used to predict the growth rate (default "fba").
    solution : cobra.Solution, optional
        A previous solution to use as a reference for (linear) MOMA or ROOM
        (default None).
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not passed,
        will be set to `configuration.processes` (default None).
    **kwargs :
        Keyword arguments are passed on to underlying simulation functions
        such as `add_room`.

    Returns
    -------
    pandas.DataFrame
        A representation of all combinations of reaction deletions. The
        columns are 'growth' and 'status', where

        index : tuple(str)
            The reaction identifiers that were knocked out.
        growth : float
            The growth rate of the adjusted model.
        status : str
            The solution's status.

    """
    reaction_list1, reaction_list2 = _element_lists(
        model.reactions, reaction_list1, reaction_list2
    )
    return _multi_deletion(
        model,
        "reaction",
        element_lists=[reaction_list1, reaction_list2],
        method=method,
        solution=solution,
        processes=processes,
        **kwargs,
    )


def double_gene_deletion(
    model: Model,
    gene_list1: Optional[List[Union[Gene, str]]] = None,
    gene_list2: Optional[List[Union[Gene, str]]] = None,
    method: str = "fba",
    solution: Optional["Solution"] = None,
    processes: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """Knock out each gene pair from the combination of two given lists.

    We say 'pair' here but the order order does not matter.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to perform deletions in.
    gene_list1 : list of cobra.Gene or str, optional
        The first gene list to be deleted. If not passed,
        all the genes from the model are used (default None).
    gene_list2 : list of cobra.Gene or str, optional
        The second gene list to be deleted. If not passed,
        all the genes from the model are used (default None).
    method : {"fba", "moma", "linear moma", "room", "linear room"}, optional
        Method used to predict the growth rate (default None).
    solution : cobra.Solution, optional
        A previous solution to use as a reference for (linear) MOMA or ROOM
        (default None).
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not passed,
        will be set to `configuration.processes` (default None).
    **kwargs :
        Keyword arguments are passed on to underlying simulation functions
        such as `add_room`.

    Returns
    -------
    pandas.DataFrame
        A representation of all combinations of gene deletions. The
        columns are 'growth' and 'status', where

        index : tuple(str)
            The gene identifiers that were knocked out.
        growth : float
            The growth rate of the adjusted model.
        status : str
            The solution's status.

    """
    gene_list1, gene_list2 = _element_lists(model.genes, gene_list1, gene_list2)
    return _multi_deletion(
        model,
        "gene",
        element_lists=[gene_list1, gene_list2],
        method=method,
        solution=solution,
        processes=processes,
        **kwargs,
    )


@pd.api.extensions.register_dataframe_accessor("knockout")
class KnockoutAccessor:
    """
    Access unique combinations of reactions in deletion results.

    This allows acces in the form of `results.knockout[rxn1]` or
    `results.knockout["rxn1_id"]`. Each individual entry will return a
    deletion so `results.knockout[rxn1, rxn2]` will return two deletions
    (for individual knockouts of rxn1 and rxn2 respectively).
    Multi-deletions can be accessed by passing in sets like
    `results.knockout[{rxn1, rxn2}]` which denotes the double deletion of
    both reactions. Thus, the following are allowed index elements:

    - single reactions or genes (depending on whether it is a gene or
      reaction deletion)
    - single reaction IDs or gene IDs
    - lists of single single reaction IDs or gene IDs (will return one row
      for each
      element in the list)
    - sets of reactions or genes (for multi-deletions)
    - sets of reactions IDs or gene IDs
    - list of sets of objects or IDs (to get several multi-deletions)

    Parameters:
    -----------
    pandas_obj : pandas.DataFrame or pandas.Series
        A result from one of the deletion methods.

    """

    def __init__(self, pandas_obj: Union[pd.DataFrame, pd.Series]) -> None:
        """Set up the accessor."""
        self._validate(pandas_obj)
        self._result = pandas_obj

    @staticmethod
    def _validate(obj: pd.DataFrame) -> None:
        """Validate the object given.

        Parameters
        ----------
        obj : pandas.DataFrame
            The object to validate.

        Raises
        ------
        AttributeError
            If the object supplied is not a DataFrame.

        """
        # verify it is a deletion results
        if any(name not in obj.columns for name in ["ids", "growth", "status"]):
            raise AttributeError("Must be DataFrame returned by a deletion method.")

    def __getitem__(
        self,
        args: Union[
            Gene,
            List[Gene],
            Set[Gene],
            List[Set[Gene]],
            Reaction,
            List[Reaction],
            Set[Reaction],
            List[Set[Reaction]],
            str,
            List[str],
            Set[str],
            List[Set[str]],
        ],
    ) -> pd.DataFrame:
        """Return the deletion result for a particular set of knocked entities.

        Parameters
        ----------
        args : cobra.Reaction, cobra.Gene, str, set, or list
            The deletions to be returned. Accepts:
            - single reactions or genes
            - single reaction IDs or gene IDs
            - lists of single single reaction IDs or gene IDs
            - sets of reactions or genes
            - sets of reactions IDs or gene IDs
            - list of sets of objects or IDs
            See the docs for usage examples.

        Returns
        -------
        pandas.DataFrame
            The deletion result where the chosen entities have been deleted.
            Each row denotes a deletion.

        Raises
        ------
        ValueError
            If any other object is used as index for lookup.

        """
        if not any(isinstance(args, t) for t in [tuple, list]):
            args = [args]

        if any(isinstance(args[0], t) for t in [Reaction, Gene, str]):
            try:
                args = [{obj.id} for obj in args]
            except AttributeError:
                # are already strings
                args = [{obj} for obj in args]
        elif isinstance(args[0], set):
            try:
                args = [{elem.id for elem in obj} for obj in args]
            except AttributeError:
                args = [set(obj) for obj in args]
        else:
            raise ValueError(
                "Allowed indices are single cobra.Reaction or cobra.Gene, "
                "lists of cobra.Reaction of cobra.Gene, or lists of sets "
                "of cobra.Reaction or cobra.Gene."
            )
        found = [x in args for x in self._result.ids]
        return self._result[found]



"""Define the Reaction class."""

import hashlib
import re
from collections import defaultdict
from copy import copy, deepcopy
from functools import partial
from math import isinf
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    AnyStr,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from warnings import warn


if TYPE_CHECKING:
    from optlang.interface import Variable
    from cobra import Model, Solution
    from cobra.summary import ReactionSummary
    import pandas as pd

from ..exceptions import OptimizationError
from ..manipulation import remove_genes
from ..util.context import get_context, resettable
from ..util.solver import (
    check_solver_status,
    linear_reaction_coefficients,
    set_objective,
)
from ..util.util import format_long_string
from .configuration import Configuration
from .gene import GPR, Gene
from .metabolite import Metabolite
from .object import Object


config = Configuration()

# This regular expression finds any single letter compartment enclosed in
# square brackets at the beginning of the string. For example [c] : foo --> bar
compartment_finder = re.compile(r"^\s*(\[[A-Za-z]\])\s*:*")
# Regular expressions to match the arrows
_reversible_arrow_finder = re.compile("<(-+|=+)>")
_forward_arrow_finder = re.compile("(-+|=+)>")
_reverse_arrow_finder = re.compile("<(-+|=+)")


class Reaction(Object):
    """Define the cobra.Reaction class.

    Reaction is a class for holding information regarding
    a biochemical reaction in a cobra.Model object.

    Reactions are by default irreversible with bounds
    `(0.0, cobra.Configuration().upper_bound)`
    if no bounds are provided on creation.
    To create an irreversible reaction use `lower_bound=None`,
    resulting in reaction bounds of
    `(cobra.Configuration().lower_bound, cobra.Configuration().upper_bound)`.

    Parameters
    ----------
    id : str, optional
        The identifier to associate with this reaction (default None).
    name : str, optional
        A human readable name for the reaction (default "").
    subsystem : str, optional
        Subsystem where the reaction is meant to occur (default "").
    lower_bound : float
        The lower flux bound (default 0.0).
    upper_bound : float, optional
        The upper flux bound (default None).
    **kwargs:
        Further keyword arguments are passed on to the parent class.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        id: Optional[str] = None,
        name: str = "",
        subsystem: str = "",
        lower_bound: float = 0.0,
        upper_bound: Optional[float] = None,
        **kwargs,
    ) -> None:
        """Initialize the Reaction class."""
        super().__init__(id, name, **kwargs)
        self._gpr = GPR()
        self.subsystem = subsystem

        # The cobra.Genes that are used to catalyze the reaction
        self._genes = set()

        # A dictionary of metabolites and their stoichiometric coefficients in
        # this reaction.
        self._metabolites = {}

        # self.model is None or refers to the cobra.Model that
        # contains self
        self._model = None

        # from cameo ...
        self._lower_bound = (
            lower_bound if lower_bound is not None else config.lower_bound
        )
        self._upper_bound = (
            upper_bound if upper_bound is not None else config.upper_bound
        )

    def _set_id_with_model(self, value: str) -> None:
        """Set Reaction id in model, check that it doesn't already exist.

        The function will rebuild the model reaction index.

        Parameters
        ----------
        value: str
            A string that represents the id.

        Raises
        ------
        ValueError
            If the model already contains a reaction with the id value.
        """
        if value in self.model.reactions:
            raise ValueError(
                f"The model already contains a reaction with the id: {value}"
            )
        forward_variable = self.forward_variable
        reverse_variable = self.reverse_variable
        self._id = value
        self.model.reactions._generate_index()
        forward_variable.name = self.id
        reverse_variable.name = self.reverse_id

    @property
    def reverse_id(self) -> str:
        """Generate the id of reverse_variable from the reaction's id.

        Returns
        -------
        str
            The original id, joined to the word reverse and a partial hash
            of the utf-8 encoded id.
        """
        return "_".join(
            (self.id, "reverse", hashlib.md5(self.id.encode("utf-8")).hexdigest()[0:5])
        )

    @property
    def flux_expression(self) -> Optional["Variable"]:
        """Get Forward flux expression.

        Returns
        -------
        flux_expression: optlang.interface.Variable, optional
            The expression representing the the forward flux (if associated
            with model), otherwise None. Representing the net flux if
            model.reversible_encoding == 'unsplit' or None if reaction is
            not associated with a model
        """
        if self.model is not None:
            return 1.0 * self.forward_variable - 1.0 * self.reverse_variable
        else:
            return None

    @property
    def forward_variable(self) -> Optional["Variable"]:
        """Get an optlang variable representing the forward flux.

        Returns
        -------
        optlang.interface.Variable, optional
            An optlang variable for the forward flux or None if reaction is
            not associated with a model.
        """
        if self.model is not None:
            return self.model.variables[self.id]
        else:
            return None

    @property
    def reverse_variable(self) -> Optional["Variable"]:
        """Get an optlang variable representing the reverse flux.

        Returns
        -------
        optlang.interface.Variable, optional
            An optlang variable for the reverse flux or None if reaction is
            not associated with a model.
        """
        if self.model is not None:
            return self.model.variables[self.reverse_id]
        else:
            return None

    @property
    def objective_coefficient(self) -> float:
        """Get the coefficient for this reaction in a linear objective (float).

        Assuming that the objective of the associated model is summation of
        fluxes from a set of reactions, the coefficient for each reaction
        can be obtained individually using this property. A more general way
        is to use the `model.objective` property directly.

        Returns
        ------
        float
            Linear coefficient if this reaction has any, or 0.0 otherwise.

        Raises
        ------
        AttributeError
            If the model of the reaction is missing (None).
        """
        return linear_reaction_coefficients(self.model, [self]).get(self, 0.0)

    @objective_coefficient.setter
    def objective_coefficient(self, value: float) -> None:
        """Set Objective coefficient.

        Parameters
        ----------
        value: float
            Number to set coefficient to.

        Raises
        ------
        AttributeError
            If reaction does not have model.
        """
        if self.model is None:
            raise AttributeError("cannot assign objective to a missing model")
        if self.flux_expression is not None:
            set_objective(self.model, {self: value}, additive=True)

    def __copy__(self) -> "Reaction":
        """Copy the Reaction.

        Returns
        -------
        Reaction
            A new reaction that is a copy of the original reaction.
        """
        cop = copy(super(Reaction, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "Reaction":
        """Copy the reaction with memo.

        Parameters
        ----------
        memo: dict
            Automatically passed parameter.

        Returns
        -------
        Reaction
            A new reaction that is a deep copy of the original reaction with memo.
        """
        cop = deepcopy(super(Reaction, self), memo)
        return cop

    @staticmethod
    def _check_bounds(lb: float, ub: float) -> None:
        """Check if the lower and upper bounds are valid.

        Parameters
        ----------
        lb: float
            The lower bound.
        ub: float
            The upper bound.

        Raises
        ------
        ValueError
            If the lower bound is higher than upper bound.
        """
        if lb > ub:
            raise ValueError(
                f"The lower bound must be less than or equal to the upper bound "
                f"({lb} <= {ub})."
            )

    def update_variable_bounds(self) -> None:
        """Update and correct variable bounds.

        Sets the forward_variable and reverse_variable bounds based on lower and
        upper bounds. This function corrects for bounds defined as inf or -inf.
        This function will also adjust the associated optlang variables associated
        with the reaction.

        See Also
        -------
        optlang.interface.set_bounds
        """
        if self.model is None:
            return
        # We know that `lb <= ub`.
        if self._lower_bound > 0:
            self.forward_variable.set_bounds(
                lb=None if isinf(self._lower_bound) else self._lower_bound,
                ub=None if isinf(self._upper_bound) else self._upper_bound,
            )
            self.reverse_variable.set_bounds(lb=0, ub=0)
        elif self._upper_bound < 0:
            self.forward_variable.set_bounds(lb=0, ub=0)
            self.reverse_variable.set_bounds(
                lb=None if isinf(self._upper_bound) else -self._upper_bound,
                ub=None if isinf(self._lower_bound) else -self._lower_bound,
            )
        else:
            self.forward_variable.set_bounds(
                lb=0, ub=None if isinf(self._upper_bound) else self._upper_bound
            )
            self.reverse_variable.set_bounds(
                lb=0, ub=None if isinf(self._lower_bound) else -self._lower_bound
            )

    @property
    def lower_bound(self) -> float:
        """Get the lower bound.

        Returns
        -------
        float
            The lower bound of the reaction.
        """
        return self._lower_bound

    @lower_bound.setter
    @resettable
    def lower_bound(self, value: float) -> None:
        """Set the lower bound.

        Parameters
        ----------
        value: float
            The value to set the lower bound.

        Setting the lower bound (float) will also adjust the associated optlang
        variables associated with the reaction.

        When using a `HistoryManager` context, this attribute can be set
        temporarily, reversed when the exiting the context.

        Raises
        ------
        ValueError
            If lower bound higher than the current upper bound. via _check_bounds.

        See Also
        --------
        _check_bounds
        """
        # Validate bounds before setting them.
        self._check_bounds(value, self._upper_bound)
        self._lower_bound = value
        self.update_variable_bounds()

    @property
    def upper_bound(self) -> float:
        """Get the upper bound.

        Returns
        -------
        float
            The upper bound of the reaction.
        """
        return self._upper_bound

    @upper_bound.setter
    @resettable
    def upper_bound(self, value: float) -> None:
        """Set the upper bound.

        Parameters
        ----------
        value: float
            The value to set the upper bound.

        Setting the upper bound (float) will also adjust the associated optlang
        variables associated with the reaction.

        When using a `HistoryManager` context, this attribute can be set
        temporarily, reversed when the exiting the context.

        Raises
        ------
        ValueError
            If upper bound lower than the current upper bound. via _check_bounds.

        See Also
        --------
        _check_bounds
        """
        # Validate bounds before setting them.
        self._check_bounds(self._lower_bound, value)
        self._upper_bound = value
        self.update_variable_bounds()

    @property
    def bounds(self) -> Tuple[float, float]:
        """Get or the bounds.

        Returns
        -------
        tuple: lower_bound, upper_bound
            A tuple of floats, representing the lower and upper bound.
        """
        return self.lower_bound, self.upper_bound

    @bounds.setter
    @resettable
    def bounds(self, value: Union[Tuple[float, float], Sequence[float]]) -> None:
        """Set the bounds directly, using a tuple or list.

        Parameters
        ----------
        value: tuple or sequence
            The lower bound and upper bound. Invalid bounds will raise ValueError.

        When using a `HistoryManager` context, this attribute can be set
        temporarily, reversed when the exiting the context.

        Raises
        ------
        ValueError
            If lower bound higher than upper bound, via _check_bounds.

        """
        lower, upper = value
        # Validate bounds before setting them.
        self._check_bounds(lower, upper)
        self._lower_bound = lower
        self._upper_bound = upper
        self.update_variable_bounds()

    @property
    def flux(self) -> float:
        """
        Get the flux value in the most recent solution.

        Flux is the primal value of the corresponding variable in the model.

        Returns
        -------
        flux: float
            Flux is the primal value of the corresponding variable in the model.

        Warnings
        --------
        * Accessing reaction fluxes through a `Solution` object is the safer,
          preferred, and only guaranteed to be correct way. You can see how to
          do so easily in the examples.
        * Reaction flux is retrieved from the currently defined
          `self._model.solver`. The solver status is checked but there are no
          guarantees that the current solver state is the one you are looking
          for.
        * If you modify the underlying model after an optimization, you will
          retrieve the old optimization values.

        Raises
        ------
        RuntimeError
            If the underlying model was never optimized beforehand or the
            reaction is not part of a model.
        OptimizationError
            If the solver status is anything other than 'optimal'.
        AssertionError
            If the flux value is not within the bounds.

        Examples
        --------
        >>> from cobra.io import load_model
        >>> model = load_model("textbook")
        >>> solution = model.optimize()
        >>> model.reactions.PFK.flux
        7.477381962160283
        >>> solution.fluxes.PFK
        7.4773819621602833
        """
        try:
            check_solver_status(self._model.solver.status)
            return self.forward_variable.primal - self.reverse_variable.primal
        except AttributeError:
            raise RuntimeError(f"reaction '{self.id}' is not part of a model")
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise err
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise OptimizationError(
                f"Likely no solution exists. Original solver message: {str(err)}."
            ) from err

    @property
    def reduced_cost(self) -> float:
        """
        Get the reduced cost in the most recent solution.

        Reduced cost is the dual value of the corresponding variable in the
        model.

        Returns
        -------
        reducd_cost: float
            A float representing the reduced cost.

        Warnings
        --------
        * Accessing reduced costs through a `Solution` object is the safer,
          preferred, and only guaranteed to be correct way. You can see how to
          do so easily in the examples.
        * Reduced cost is retrieved from the currently defined
          `self._model.solver`. The solver status is checked but there are no
          guarantees that the current solver state is the one you are looking
          for.
        * If you modify the underlying model after an optimization, you will
          retrieve the old optimization values.

        Raises
        ------
        RuntimeError
            If the underlying model was never optimized beforehand or the
            reaction is not part of a model.
        OptimizationError
            If the solver status is anything other than 'optimal'.

        Examples
        --------
        >>> from cobra.io import load_model
        >>> model = load_model("textbook")
        >>> solution = model.optimize()
        >>> model.reactions.PFK.reduced_cost
        -8.673617379884035e-18
        >>> solution.reduced_costs.PFK
        -8.6736173798840355e-18
        """
        try:
            check_solver_status(self._model.solver.status)
            return self.forward_variable.dual - self.reverse_variable.dual
        except AttributeError:
            raise RuntimeError(f"reaction '{self.id}' is not part of a model")
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise err
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise OptimizationError(
                f"Likely no solution exists. Original solver message: {str(err)}."
            ) from err

    # read-only
    @property
    def metabolites(self) -> Dict[Metabolite, float]:
        """Get a dictionary of metabolites and coefficients.

        Returns
        -------
        metaoblites: Dict[Metabolite, float]
            A copy of self._metabolites, which is a dictionary of cobra.Metabolite for
            keys and floats for coeffecieints. Positive coefficient means the reaction
            produces this metabolite, while negative coefficient means the reaction
            consumes this metabolite.
        """
        return self._metabolites.copy()

    @property
    def genes(self) -> FrozenSet:
        """Return the genes of the reaction.

        Returns
        -------
        genes: FrozenSet
        """
        return frozenset(self._genes)

    def update_genes_from_gpr(self) -> None:
        """Update genes of reation based on GPR.

        If the reaction has a model, and new genes appear in the GPR, they will be
        created as Gene() entities and added to the model. If the reaction doesn't have
        a model, genes will be created without a model.

        Genes that no longer appear in the GPR will be removed from the reaction, but
        not the model. If you want to remove them expliclty, use model.remove_genes().
        """
        context = get_context(self)
        if self._gpr.body is not None:
            new_gene_names = self._gpr.genes
        else:
            new_gene_names = set()
        old_genes = self._genes.copy()
        new_genes = set()
        if self._model is None:
            self._genes = {Gene(i) for i in new_gene_names}
        else:
            model_genes = self._model.genes
            self._genes = set()
            for g_id in new_gene_names:
                if not model_genes.has_id(g_id):
                    new_gene = Gene(g_id)
                    new_gene._model = self._model
                    model_genes.append(new_gene)
                    if context:
                        # Remove the gene later
                        context(
                            partial(
                                remove_genes,
                                model=self._model,
                                gene_list=[model_genes.get_by_id(g_id)],
                                remove_reactions=False,
                            )
                        )
                        context(partial(setattr, new_gene, "_model", None))
                        # Maybe should be
                        # context(partial(self._model.genes.__isub__, [new_gene]))
                new_gene = model_genes.get_by_id(g_id)
                self._genes.add(new_gene)
                new_genes.add(new_gene)

        # Make the genes aware that it is involved in this reaction
        for g in self._genes:
            self._associate_gene(g)
            if context:
                context(partial(self._dissociate_gene, g))

        # make the old genes aware they are no longer involved in this reaction
        for g in old_genes.difference(new_genes):
            try:
                self._dissociate_gene(g)
                if context:
                    context(partial(self._associate_gene, g))
            except KeyError:
                warn(f"could not remove old gene {g.id} from reaction {self.id}")
            if g in self._genes:  # if an old gene is still a new gene
                raise Exception("something wrong with sets. Shouldn't happen.")

    @property
    def gene_reaction_rule(self) -> str:
        """See gene reaction rule as str.

        Uses the to_string() method of the GPR class

        Returns
        -------
        str

        """
        return self._gpr.to_string()

    @gene_reaction_rule.setter
    @resettable
    def gene_reaction_rule(self, new_rule: str) -> None:
        """Set a new GPR for the reaction, using a str expression.

        Will use the new GPR to update reaction genes.

        Parameters
        ----------
        new_rule : str
            which will be parsed by the string parser in GPR, GPR.from_string(new_rule).
            It makes a new GPR, and does not modify the existing one.

        See Also
        --------
        update_genes_from_gpr()

        """
        self._gpr = GPR.from_string(new_rule)
        self.update_genes_from_gpr()

    @property
    def gene_name_reaction_rule(self):
        """Display gene_reaction_rule with names intead.

        Do NOT use this string for computation. It is intended to give a
        representation of the rule using more familiar gene names instead of
        the often cryptic ids.

        """
        names = {i.id: i.name for i in self._genes}
        return self._gpr.to_string(names=names)

    @property
    def gpr(self) -> GPR:
        """Return the GPR associated with the reaction.

        Returns
        -------
        gpr: GPR
            The GPR class, see cobra.core.gene.GPR() for details.
        """
        return self._gpr

    @gpr.setter
    @resettable
    def gpr(self, value: GPR) -> None:
        """Set a new GPR for the reaction, using GPR() class.

        Also updates the reaction genes based on GPR.

        Parameters
        ----------
        value : GPR() class.

        See Also
        --------
        cobra.core.gene.GPR()
        update_genes_from_gpr()
        """
        self._gpr = value
        self.update_genes_from_gpr()

    @property
    def functional(self) -> bool:
        """All required enzymes for reaction are functional.

        Returns
        -------
        bool
            True if the gene-protein-reaction (GPR) rule is fulfilled for
            this reaction, or if reaction is not associated to a model,
            otherwise False.
        """
        if self._model:
            return self._gpr.eval(
                {gene.id for gene in self.genes if not gene.functional}
            )
        return True

    @property
    def x(self) -> float:
        """Get the flux through the reaction in the most recent solution.

        Flux values are computed from the primal values of the variables in
        the solution.

        Returns
        -------
        flux: float
            Float representing the flux value.

        .. deprecated ::
        Use reaction.flux instead.
        """
        warn("Please use reaction.flux instead.", DeprecationWarning)
        return self.flux

    @property
    def y(self) -> float:
        """Get the reduced cost of the reaction in the most recent solution.

        Reduced costs are computed from the dual values of the variables in
        the solution.

        Returns
        -------
        flux: float
            Float representing the reduced cost value.

        .. deprecated ::
        Use reaction.reduced_cost instead.
        """
        warn("Please use reaction.reduced_cost instead.", DeprecationWarning)
        return self.reduced_cost

    @property
    def reversibility(self) -> bool:
        """Whether the reaction can proceed in both directions (reversible).

        This is computed from the current upper and lower bounds.

        Returns
        -------
        bool:
            True if the reaction is reversible (lower bound lower than 0 and upper
            bound is higher than 0).

        """
        return self._lower_bound < 0 < self._upper_bound

    @reversibility.setter
    def reversibility(self, value: bool) -> None:
        """Set reversiblitly (ignored).

        At this stage this is not set and is ignored. Present for compatiblity reasons.
        If you'd like to change the reversibility of the reaction change
        lower and upper bounds.

        Parameters
        ----------
        value: bool
            This value is ignored.
        """
        warn("Setting reaction reversibility is ignored")

    @property
    def boundary(self) -> bool:
        """Whether or not this reaction is an exchange reaction.

        Returns
        -------
        bool:   `True` if the reaction has either no products or reactants.
        """
        return len(self.metabolites) == 1 and not (self.reactants and self.products)

    @property
    def model(self) -> Optional["Model"]:
        """Return the model the reaction is a part of.

        Returns
        -------
        model: cobra.Model, optional
            The model this reaction belongs to. None if there is no model associated
            with this reaction.
        """
        return self._model

    def _update_awareness(self) -> None:
        """Update awareness for genes and metaoblites of the reaction.

        Make sure all metabolites and genes that are associated with
        this reaction are aware of it.

        """
        for x in self._metabolites:
            x._reaction.add(self)
        for x in self._genes:
            x._reaction.add(self)

    def remove_from_model(self, remove_orphans: bool = False) -> None:
        """Remove the reaction from a model.

        This removes all associations between a reaction the associated
        model, metabolites and genes.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        remove_orphans : bool
            Remove orphaned genes and metabolites from the model as well (default
            False).
        """
        self._model.remove_reactions([self], remove_orphans=remove_orphans)

    def delete(self, remove_orphans: bool = False) -> None:
        """Remove the reaction from a model.

        This removes all associations between a reaction the associated
        model, metabolites and genes.

        The change is reverted upon exit when using the model as a context.

        .. deprecated ::
        use `reaction.remove_from_model` instead.

        Parameters
        ----------
        remove_orphans : bool
            Remove orphaned genes and metabolites from the model as well (default
            False).
        """
        warn(
            "delete is deprecated. Use reaction.remove_from_model instead",
            DeprecationWarning,
        )
        self.remove_from_model(remove_orphans=remove_orphans)

    def __getstate__(self) -> Dict:
        """Get state for reaction.

        This serializes the reaction object. The GPR will be converted to a string
        to avoid unneccessary copies due to interdependencies of used objects.

        Returns
        -------
        dict
            The state/attributes of the reaction in serilized form.

        """
        state = self.__dict__.copy()
        state["_gpr"] = str(self._gpr)
        return state

    def __setstate__(self, state: Dict) -> None:
        """Set state for reaction.

        Probably not necessary to set _model as the cobra.Model that
        contains self sets the _model attribute for all metabolites and genes
        in the reaction.

        However, to increase performance speed we do want to let the metabolite
        and gene know that they are employed in this reaction

        Parameters
        ----------
        state: dict
            A dictionary of state, where keys are attribute names (str).
            Similar to __dict__.
        """
        # These are necessary for old pickles which store attributes
        # which have since been superceded by properties.
        if "reaction" in state:
            state.pop("reaction")
        if "gene_reaction_rule" in state:
            state["_gene_reaction_rule"] = state.pop("gene_reaction_rule")
        if "lower_bound" in state:
            state["_lower_bound"] = state.pop("lower_bound")
        if "upper_bound" in state:
            state["_upper_bound"] = state.pop("upper_bound")

        # Used for efficient storage in newer cobrapy versions
        if type(state["_gpr"]) is str:
            state["_gpr"] = GPR.from_string(state["_gpr"])

        self.__dict__.update(state)
        for x in state["_metabolites"]:
            x._model = self._model
            x._reaction.add(self)
        for x in state["_genes"]:
            x._model = self._model
            x._reaction.add(self)

    def copy(self) -> "Reaction":
        """Copy a reaction.

        The referenced metabolites and genes are also copied.

        Returns
        -------
        cobra.Reaction
            A copy of the Reaction.
        """
        # no references to model when copying
        model = self._model
        self._model = None
        for i in self._metabolites:
            i._model = None
        for i in self._genes:
            i._model = None
        # now we can copy
        new_reaction = deepcopy(self)
        # restore the references
        self._model = model
        for i in self._metabolites:
            i._model = model
        for i in self._genes:
            i._model = model
        return new_reaction

    def __add__(self, other: "Reaction") -> "Reaction":
        """Add two reactions and return a new one.

        The stoichiometry will be the combined stoichiometry of the two
        reactions, and the gene reaction rule will be both rules combined by an
        and. All other attributes (i.e. reaction bounds) will match those of
        the first reaction.

        Does not modify in place.

        Parameters
        ----------
        other: cobra.Reaction
            Another reaction to add to the current one.

        Returns
        -------
        Reaction - new reaction with the added properties.

        """
        new_reaction = self.copy()
        if other == 0:
            return new_reaction
        else:
            new_reaction += other

        return new_reaction

    __radd__ = __add__

    def __iadd__(self, other: "Reaction") -> "Reaction":
        """Add two reactions in place and return the modified first one.

        The stoichiometry will be the combined stoichiometry of the two
        reactions, and the gene reaction rule will be both rules combined by an
        and. All other attributes (i.e. reaction bounds) will match those of
        the first reaction.

        Modifies in place.

        Parameters
        ----------
        other: cobra.Reaction
            Another reaction to add to the current one.

        Returns
        -------
        Reaction - original reaction (self) with the added properties.
        """
        self.add_metabolites(other._metabolites, combine=True)
        rule1 = self.gene_reaction_rule.strip()
        rule2 = other.gene_reaction_rule.strip()
        if rule1 != "" and rule2 != "":
            self.gene_reaction_rule = (
                f"({self.gene_reaction_rule}) and " f"({other.gene_reaction_rule})"
            )
        elif rule1 != "" and rule2 == "":
            self.gene_reaction_rule = rule1
        elif rule1 == "" and rule2 != "":
            self.gene_reaction_rule = rule2
        return self

    def __sub__(self, other: "Reaction") -> "Reaction":
        """Subtract two reactions and return a new one.

        The stoichiometry will be the subtracted stoichiometry of the two
        reactions, and the gene_reaction_rule will be the gene_reaction_rule of the
        first reaction. All other attributes (i.e. reaction bounds) will match those of
        the first reaction.

        Does not modify in place. The name will still be that of the first reaction.

        Parameters
        ----------
        other: Reaction
            The reaction to subtract from self.

        Returns
        -------
        Reaction - new reaction with the added properties.
        """
        new = self.copy()
        new -= other
        return new

    def __isub__(self, other: "Reaction") -> "Reaction":
        """Subtract metabolites of one reaction from another in place.

        The stoichiometry will be the metabolites of self minus the metabolites
         of the other. All other attributes including gene_reaction_rule
         (i.e. reaction bounds) will match those of
        the first reaction.

        Modifies in place and changes the original reaction.

        Parameters
        ----------
        other: Reaction
            The reaction to subtract from self.

        Returns
        -------
        Reaction - self with the subtracted metabolites.
        """
        self.subtract_metabolites(other._metabolites, combine=True)
        return self

    def __imul__(self, coefficient: float) -> "Reaction":
        """Scale coefficients in a reaction by a given value in place.

        E.g. A -> B becomes 2A -> 2B.

        If coefficient is less than zero, the reaction is reversed and the
        bounds are swapped.

        Parameters
        ----------
        coefficient: float
            Value to scale coefficients of metabolites by. If less than zero, reverses
            the reaction.

        Returns
        -------
        Reaction
            Returns the same reaction modified in place.
        """
        self._metabolites = {
            met: value * coefficient for met, value in self._metabolites.items()
        }

        if coefficient < 0:
            self.bounds = (-self.upper_bound, -self.lower_bound)

        if self._model:
            self._model._populate_solver([self])

        context = get_context(self)
        if context:
            context(partial(self._model._populate_solver, [self]))
            context(partial(self.__imul__, 1.0 / coefficient))

        return self

    def __mul__(self, coefficient: float) -> "Reaction":
        """Scale coefficients in a reaction by a given value and return new reaction.

        E.g. A -> B becomes 2A -> 2B.

        If coefficient is less than zero, the reaction is reversed and the
        bounds are swapped.

        Parameters
        ----------
        coefficient: float
            Value to scale coefficients of metabolites by. If less than zero, reverses
            the reaction.

        Returns
        -------
        Reaction
            Returns a new reaction, identical to the original except coefficients.
        """
        new = self.copy()
        new *= coefficient
        return new

    @property
    def reactants(self) -> List[Metabolite]:
        """Return a list of reactants for the reaction.

        Returns
        -------
        list
            A list of the metabolites consudmed (coefficient < 0) by the reaction.
        """
        return [k for k, v in self._metabolites.items() if v < 0]

    @property
    def products(self) -> List[Metabolite]:
        """Return a list of products for the reaction.

        Returns
        -------
        list
            A list of the metabolites produced (coefficient > 0) by the reaction.
        """
        return [k for k, v in self._metabolites.items() if v >= 0]

    def get_coefficient(self, metabolite_id: Union[str, Metabolite]) -> float:
        """Return the stoichiometric coefficient of a metabolite.

        Parameters
        ----------
        metabolite_id : str or cobra.Metabolite

        """
        if isinstance(metabolite_id, Metabolite):
            return self._metabolites[metabolite_id]

        _id_to_metabolites = {m.id: m for m in self._metabolites}
        return self._metabolites[_id_to_metabolites[metabolite_id]]

    def get_coefficients(
        self, metabolite_ids: Iterable[Union[str, Metabolite]]
    ) -> Iterator[float]:
        """Return the stoichiometric coefficients for a list of metabolites.

        Parameters
        ----------
        metabolite_ids : iterable
            Containing ``str`` or ``cobra.Metabolite``s.

        Returns
        -------
        map: Iterable
            Returns the result of map function, which is a map object (an Iterable).

        """
        return map(self.get_coefficient, metabolite_ids)

    def add_metabolites(
        self,
        metabolites_to_add: Dict[Metabolite, float],
        combine: bool = True,
        reversibly: bool = True,
    ) -> None:
        """Add metabolites and stoichiometric coefficients to the reaction.

        If the final coefficient for a metabolite is 0 then it is removed
        from the reaction.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        metabolites_to_add : dict
            Dictionary with metabolite objects or metabolite identifiers as
            keys and coefficients as values. If keys are strings (name of a
            metabolite) the reaction must already be part of a model and a
            metabolite with the given name must exist in the model.

        combine : bool
            Describes behavior if a metabolite already exists in the reaction (default
            True).
            True causes the coefficients to be added.
            False causes the coefficient to be replaced.

        reversibly : bool
            Whether to add the change to the context to make the change
            reversibly or not (primarily intended for internal use). Default is True.

        Raises
        ------
        KeyError
            If the metabolite string id is not in the model.
        ValueError
            If the metabolite key in the dictionary is a string, and there is no model
            for the reaction.
        """
        old_coefficients = self.metabolites
        new_metabolites = []
        _id_to_metabolites = dict([(x.id, x) for x in self._metabolites])

        for metabolite, coefficient in metabolites_to_add.items():
            # Make sure metabolites being added belong to the same model, or
            # else copy them.
            if isinstance(metabolite, Metabolite):
                if (metabolite.model is not None) and (
                    metabolite.model is not self._model
                ):
                    metabolite = metabolite.copy()

            met_id = str(metabolite)
            # If a metabolite already exists in the reaction then
            # just add them.
            if met_id in _id_to_metabolites:
                reaction_metabolite = _id_to_metabolites[met_id]
                if combine:
                    self._metabolites[reaction_metabolite] += coefficient
                else:
                    self._metabolites[reaction_metabolite] = coefficient
            else:
                # If the reaction is in a model, ensure we aren't using
                # a duplicate metabolite.
                if self._model:
                    try:
                        metabolite = self._model.metabolites.get_by_id(met_id)
                    except KeyError as e:
                        if isinstance(metabolite, Metabolite):
                            new_metabolites.append(metabolite)
                        else:
                            # do we want to handle creation here?
                            raise e
                elif isinstance(metabolite, str):
                    # if we want to handle creation, this should be changed
                    raise ValueError(
                        f"Reaction '{self.id}' does not belong to a model. "
                        f"Either add the reaction to a model or use Metabolite objects "
                        f"instead of strings as keys."
                    )
                self._metabolites[metabolite] = coefficient
                # make the metabolite aware that it is involved in this
                # reaction
                metabolite._reaction.add(self)

        # from cameo ...
        model = self.model
        if model is not None:
            model.add_metabolites(new_metabolites)

            for metabolite, coefficient in self._metabolites.items():
                model.constraints[metabolite.id].set_linear_coefficients(
                    {
                        self.forward_variable: coefficient,
                        self.reverse_variable: -coefficient,
                    }
                )

        for metabolite, the_coefficient in list(self._metabolites.items()):
            if the_coefficient == 0:
                # make the metabolite aware that it no longer participates
                # in this reaction
                metabolite._reaction.remove(self)
                self._metabolites.pop(metabolite)

        context = get_context(self)
        if context and reversibly:
            if combine:
                # Just subtract the metabolites that were added
                context(
                    partial(
                        self.subtract_metabolites,
                        metabolites_to_add,
                        combine=True,
                        reversibly=False,
                    )
                )
            else:
                # Reset them with add_metabolites
                mets_to_reset = {
                    key: old_coefficients[model.metabolites.get_by_any(key)[0]]
                    for key in metabolites_to_add.keys()
                }

                context(
                    partial(
                        self.add_metabolites,
                        mets_to_reset,
                        combine=False,
                        reversibly=False,
                    )
                )

    def subtract_metabolites(
        self,
        metabolites: Dict[Metabolite, float],
        combine: bool = True,
        reversibly: bool = True,
    ) -> None:
        """Subtract metabolites from a reaction.

        That means add the metabolites with -1*coefficient. If the final
        coefficient for a metabolite is 0 then the metabolite is removed from
        the reaction.

        Notes
        -----
        * A final coefficient < 0 implies a reactant.
        * The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        metabolites : dict
            Dictionary where the keys are of class Metabolite and the values
            are the coefficients. These metabolites will be added to the
            reaction.

        combine : bool
            Describes behavior if a metabolite already exists in the reaction (default
            True).
            True causes the coefficients to be added.
            False causes the coefficient to be replaced.

        reversibly : bool
            Whether to add the change to the context to make the change
            reversibly or not ,primarily intended for internal use (default True).
        """
        self.add_metabolites(
            {k: -v for k, v in metabolites.items()},
            combine=combine,
            reversibly=reversibly,
        )

    @property
    def reaction(self) -> str:
        """Return Human readable reaction str.

        Returns
        -------
        str
            The reaction in a human readble str.
        """
        return self.build_reaction_string()

    @reaction.setter
    def reaction(self, value: str) -> None:
        """Set reaction from Human readable reaction str.

        Parameters
        -------
        value: str
            The reaction in a human readble str.

        See Also
        --------
        build_reaction_from_string()
        """
        self.build_reaction_from_string(value)

    def build_reaction_string(self, use_metabolite_names: bool = False) -> str:
        """Generate a human readable reaction str.

        Parameters
        ----------
        use_metabolite_names: bool
            Whether to use metabolite names (when True) or metabolite ids (when False,
            default).

        Returns
        -------
        str
            A human readable str.
        """

        def format(number: Union[int, float]) -> str:
            return "" if number == 1 else str(number).rstrip(".") + " "

        id_type = "id"
        if use_metabolite_names:
            id_type = "name"
        reactant_bits = []
        product_bits = []
        for met in sorted(self._metabolites, key=attrgetter("id")):
            coefficient = self._metabolites[met]
            name = str(getattr(met, id_type))
            if coefficient >= 0:
                product_bits.append(format(coefficient) + name)
            else:
                reactant_bits.append(format(abs(coefficient)) + name)

        reaction_string = " + ".join(reactant_bits)
        if not self.reversibility:
            if self.lower_bound < 0 and self.upper_bound <= 0:
                reaction_string += " <-- "
            else:
                reaction_string += " --> "
        else:
            reaction_string += " <=> "
        reaction_string += " + ".join(product_bits)
        return reaction_string

    def check_mass_balance(self) -> Dict[str, float]:
        """Compute mass and charge balance for the reaction.

        Returns
        -------
        dict
            a dict of {element: amount} for unbalanced elements.
            "charge" is treated as an element in this dict
            This should be empty for balanced reactions.

        Raises
        ------
        ValueError
            No elements were found in metabolite.
        """
        reaction_element_dict = defaultdict(int)
        for metabolite, coefficient in self._metabolites.items():
            if metabolite.charge is not None:
                reaction_element_dict["charge"] += coefficient * metabolite.charge
            if metabolite.elements is None:
                raise ValueError(f"No elements found in metabolite {metabolite.id}")
            for element, amount in metabolite.elements.items():
                reaction_element_dict[element] += coefficient * amount
        # filter out 0 values
        return {k: v for k, v in reaction_element_dict.items() if v != 0}

    @property
    def compartments(self) -> Set:
        """Return set of compartments the metabolites are in.

        Returns
        -------
        set
            A set of compartments the metabolites are in.
        """
        return {
            met.compartment for met in self._metabolites if met.compartment is not None
        }

    def get_compartments(self) -> list:
        """List compartments the metabolites are in.

        Returns
        -------
        list
            A list of compartments the metabolites are in.
        .. deprecated ::
        Use reaction.compartments() instead.
        """
        warn("use Reaction.compartments instead", DeprecationWarning)
        return list(self.compartments)

    def _associate_gene(self, cobra_gene: Gene) -> None:
        """Associates a cobra.Gene object with a cobra.Reaction.

        Parameters
        ----------
        cobra_gene : cobra.core.Gene.Gene

        """
        self._genes.add(cobra_gene)
        cobra_gene._reaction.add(self)
        cobra_gene._model = self._model

    def _dissociate_gene(self, cobra_gene: Gene) -> None:
        """Dissociates a cobra.Gene object with a cobra.Reaction.

        Parameters
        ----------
        cobra_gene : cobra.core.Gene.Gene

        """
        self._genes.discard(cobra_gene)
        cobra_gene._reaction.discard(self)

    def knock_out(self) -> None:
        """Knockout reaction by setting its bounds to zero."""
        self.bounds = (0, 0)

    def build_reaction_from_string(
        self,
        reaction_str: str,
        verbose: bool = True,
        fwd_arrow: Optional[AnyStr] = None,
        rev_arrow: Optional[AnyStr] = None,
        reversible_arrow: Optional[AnyStr] = None,
        term_split: str = "+",
    ) -> None:
        """Build reaction from reaction equation reaction_str using parser.

        Takes a string and using the specifications supplied in the optional
        arguments infers a set of metabolites, metabolite compartments and
        stoichiometries for the reaction.  It also infers the reversibility
        of the reaction from the reaction arrow.

        Changes to the associated model are reverted upon exit when using
        the model as a context.

        Parameters
        ----------
        reaction_str : str
            a string containing a reaction formula (equation)
        verbose: bool
            setting verbosity of function (default True)
        fwd_arrow : AnyStr, optional
            Str or bytes that encode forward irreversible reaction arrows (default
            None).
        rev_arrow : AnyStr, optional
            Str or bytes that encode backward irreversible reaction arrows (default
            None).
        reversible_arrow : AnyStr, optional
            Str or bytes that encode reversible reaction arrows (default None).
        term_split : str
            dividing individual metabolite entries (default "+")".

        Raises
        ------
        ValueError
            No arrow found in reaction string.
        """
        # set the arrows
        forward_arrow_finder = (
            _forward_arrow_finder
            if fwd_arrow is None
            else re.compile(re.escape(fwd_arrow))
        )
        reverse_arrow_finder = (
            _reverse_arrow_finder
            if rev_arrow is None
            else re.compile(re.escape(rev_arrow))
        )
        reversible_arrow_finder = (
            _reversible_arrow_finder
            if reversible_arrow is None
            else re.compile(re.escape(reversible_arrow))
        )
        if self._model is None:
            warn("no model found")
            model = None
        else:
            model = self._model
        found_compartments = compartment_finder.findall(reaction_str)
        if len(found_compartments) == 1:
            compartment = found_compartments[0]
            reaction_str = compartment_finder.sub("", reaction_str)
        else:
            compartment = ""

        # reversible case
        arrow_match = reversible_arrow_finder.search(reaction_str)
        if arrow_match is not None:
            self.bounds = config.lower_bound, config.upper_bound
        else:  # irreversible
            # try forward
            arrow_match = forward_arrow_finder.search(reaction_str)
            if arrow_match is not None:
                self.bounds = 0, config.upper_bound
            else:
                # must be reverse
                arrow_match = reverse_arrow_finder.search(reaction_str)
                if arrow_match is None:
                    raise ValueError(f"no suitable arrow found in '{reaction_str}'")
                else:
                    self.bounds = config.lower_bound, 0
        reactant_str = reaction_str[: arrow_match.start()].strip()
        product_str = reaction_str[arrow_match.end() :].strip()

        self.subtract_metabolites(self.metabolites, combine=True)

        for substr, factor in ((reactant_str, -1), (product_str, 1)):
            if len(substr) == 0:
                continue
            for term in substr.split(term_split):
                term = term.strip()
                if term.lower() == "nothing":
                    continue
                if " " in term:
                    num_str, met_id = term.split()
                    num = float(num_str.lstrip("(").rstrip(")")) * factor
                else:
                    met_id = term
                    num = factor
                met_id += compartment
                try:
                    met = model.metabolites.get_by_id(met_id)
                except KeyError:
                    if verbose:
                        print(f"unknown metabolite '{met_id}' created")
                    met = Metabolite(met_id)
                self.add_metabolites({met: num})

    def summary(
        self,
        solution: Optional["Solution"] = None,
        fva: Optional[Union[float, "pd.DataFrame"]] = None,
    ) -> "ReactionSummary":
        """
        Create a summary of the reaction flux.

        Parameters
        ----------
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Returns
        -------
        cobra.summary.ReactionSummary

        See Also
        --------
        Metabolite.summary
        Model.summary

        """
        from cobra.summary import ReactionSummary

        return ReactionSummary(
            reaction=self,
            model=self._model,
            solution=solution,
            fva=fva,
        )

    def __str__(self) -> str:
        """Return reaction id and reaction as str.

        Returns
        -------
        str
            A string comprised out of reaction id and reaction.
        """
        return f"{self.id}: {self.build_reaction_string()}"

    def _repr_html_(self) -> str:
        """Generate html representation of reaction.

        Returns
        -------
        str
            HTML representation of the reaction.
        """
        return f"""
        <table>
            <tr>
                <td><strong>Reaction identifier</strong></td><td>{format_long_string(
            self.id, 100)}</td>
            </tr><tr>
                <td><strong>Name</strong></td><td>{format_long_string(
            self.name, 100)}</td>
            </tr><tr>
                <td><strong>Memory address</strong></td>
                <td>{f"{id(self):#x}"}</td>
            </tr><tr>
                <td><strong>Stoichiometry</strong></td>
                <td>
                    <p style='text-align:right'>{format_long_string(
            self.build_reaction_string(), 200)}</p>
                    <p style='text-align:right'>{format_long_string(
            self.build_reaction_string(True), 200)}</p>
                </td>
            </tr><tr>
                <td><strong>GPR</strong></td><td>{format_long_string(
            self.gene_reaction_rule, 100)}</td>
            </tr><tr>
                <td><strong>Lower bound</strong></td><td>{self.lower_bound}</td>
            </tr><tr>
                <td><strong>Upper bound</strong></td><td>{self.upper_bound}</td>
            </tr>
        </table>
        """



"""Provide functions for I/O in MATLAB (.mat) format."""

import logging
import re
from collections import OrderedDict
from pathlib import Path
from typing import IO, Dict, Iterable, List, Optional, Pattern, Union

import numpy as np

from ..core import Gene, Group, Metabolite, Model, Object, Reaction
from ..util import create_stoichiometric_matrix
from ..util.solver import set_objective


try:
    import scipy.io as scipy_io
    import scipy.sparse as scipy_sparse
except ImportError:
    scipy_sparse = None
    scipy_io = None

logger = logging.getLogger(__name__)

# The following dictionaries are based on
# https://github.com/opencobra/cobratoolbox/blob/docs/source/notes/COBRAModelFields.md
# at commit 83d26938a9babff79289d40e20f5f50dd5b710fa
# Which is the most updated master as of April 4th, 2022

MET_MATLAB_TO_PROVIDERS = {
    "metHMDBID": "hmdb",
    "metInChIString": "inchi",
    "metKEGGID": "kegg.compound",
    "metKEGGGlycanID": "kegg.glycan",
    "metKEGGDrugID": "kegg.drug",
    "metUniPathway": "unipathway.compound",
    "metPubChemID": "pubchem.compound",
    "metPubChemSubstance": "pubchem.substance",
    "metCHEBIID": "chebi",
    "metMetaNetXID": "metanetx.chemical",
    "metSEEDID": "seed.compound",
    "metBiGGID": "bigg.metabolite",
    "metBioCycID": "biocyc",
    "metEnviPathID": "envipath",
    "metLIPIDMAPSID": "lipidmaps",
    "metReactomeID": "reactome",
    "metSABIORKID": "sabiork.compound",
    "metSLMID": "slm",
    "metSMILES": "SMILES",
    "metSBOTerms": "sbo",
    "metCasNumber": "cas",
}

MET_PROVIDERS_TO_MATLAB = {
    MET_MATLAB_TO_PROVIDERS[k]: k for k in MET_MATLAB_TO_PROVIDERS.keys()
}

MET_NOTES_TO_MATLAB = {
    "metNotes": "metNotes",
}

MET_MATLAB_TO_NOTES = {MET_NOTES_TO_MATLAB[k]: k for k in MET_NOTES_TO_MATLAB.keys()}

RXN_MATLAB_TO_PROVIDERS = {
    "rxnECNumbers": "ec-code",
    "rxnReferences": "pubmed",
    "rxnKEGGID": "kegg.reaction",
    "rxnKEGGPathways": "kegg.pathway",
    "rxnMetaNetXID": "metanetx.reaction",
    "rxnSEEDID": "seed.reaction",
    "rxnBiGGID": "bigg.reaction",
    "rxnBioCycID": "biocyc",
    "rxnRheaID": "rhea",
    "rxnReactomeID": "reactome",
    "rxnSABIORKID": "sabiork.reaction",
    "rxnBRENDAID": "brenda",
    "rxnSBOTerms": "sbo",
}

RXN_PROVIDERS_TO_MATLAB = {
    RXN_MATLAB_TO_PROVIDERS[k]: k for k in RXN_MATLAB_TO_PROVIDERS.keys()
}

CONFIDENCE_STR = "Confidence Level"
RXN_MATLAB_TO_NOTES = {
    "rxnReferences": "References",
    "rxnNotes": "NOTES",
    "rxnConfidenceScores": CONFIDENCE_STR,
}

RXN_NOTES_TO_MATLAB = {
    CONFIDENCE_STR: "rxnConfidenceScores",
    "NOTES": "rxnNotes",
    "References": "rxnNotes",
}

GENE_MATLAB_TO_PROVIDERS = {
    "geneEntrezID": "ncbigene",
    "geneRefSeqID": "refseq",
    "geneUniprotID": "uniprot",
    "geneEcoGeneID": "ecogene",
    "geneKEGGID": "kegg.gene",
    "geneHPRDID": "hprd",
    "geneASAPID": "asap",
    "geneCCDSID": "ccds",
    "geneNCBIProteinID": "ncbiprotein",
}

GENE_PROVIDERS_TO_MATLAB = {
    GENE_MATLAB_TO_PROVIDERS[k]: k for k in GENE_MATLAB_TO_PROVIDERS.keys()
}

DICT_GENE = "DICT_GENE"
DICT_GENE_REV = "DICT_GENE_REV"
DICT_MET = "DICT_MET"
DICT_MET_REV = "DICT_MET_REV"
DICT_MET_NOTES = "DICT_MET_NOTES"
DICT_MET_NOTES_REV = "DICT_MET_NOTES_REV"
DICT_REACTION = "DICT_REACTION"
DICT_REACTION_REV = "DICT_REACTION_REV"
DICT_REACTION_NOTES = "DICT_REACTION_NOTES"
DICT_REACTION_NOTES_REV = "DICT_REACTION_NOTES_REV"

DICT_REPLACE: dict = {
    DICT_GENE: GENE_MATLAB_TO_PROVIDERS,
    DICT_GENE_REV: GENE_PROVIDERS_TO_MATLAB,
    DICT_MET: MET_MATLAB_TO_PROVIDERS,
    DICT_MET_REV: MET_PROVIDERS_TO_MATLAB,
    DICT_MET_NOTES: MET_MATLAB_TO_NOTES,
    DICT_MET_NOTES_REV: MET_NOTES_TO_MATLAB,
    DICT_REACTION: RXN_MATLAB_TO_PROVIDERS,
    DICT_REACTION_REV: RXN_PROVIDERS_TO_MATLAB,
    DICT_REACTION_NOTES: RXN_MATLAB_TO_NOTES,
    DICT_REACTION_NOTES_REV: RXN_NOTES_TO_MATLAB,
}

# precompiled regular expressions (kept globally for caching)
_bracket_re = re.compile(r"[\[(](?P<compartment>[a-zA-Z]+)[])]$")
# Some older models have _boundary for (some) exchange reactions
_underscore_re = re.compile(r"_(?P<compartment>[a-zA-Z]+)(_boundary)?$")
_pubmed_re = re.compile("PMID: ?(\\d+),?")
_punctuation_re = re.compile(r"^[;,.'\"]+$")
_double_punctuation_re = re.compile(r"[;,.'\"]{2,}")
_ec_re = re.compile(r"([\d\-]+.[\d\-]+.[\d\-]+.[\d-]+)")
_chebi_re = re.compile(r"\D*(\d+),?")
_sbo_re = re.compile(r"\D*(\d+),?")


def _get_id_compartment(_id: str) -> str:
    """Extract the compartment from the `id` string.

    Parameters
    ----------
    _id : str
        The ID string to extract component from.

    Returns
    -------
    str
        The extracted component string.

    """
    bracket_search = _bracket_re.search(_id)
    if bracket_search:
        return bracket_search.group("compartment")

    underscore_search = _underscore_re.search(_id)
    if underscore_search:
        return underscore_search.group("compartment")


def _cell(x: Iterable[str]) -> np.ndarray:
    """Translate an iterable `x` into a MATLAB cell array.

    Parameters
    ----------
    x : iterable of str
        The data iterable to convert to cell array.

    Returns
    -------
    numpy.ndarray
       The converted cell array compatible with MATLAB.

    """
    x_no_none = [i if i is not None else "" for i in x]
    return np.array(x_no_none, dtype=object)


def _cell_to_str_list(
    m_cell: np.ndarray,
    empty_value: Optional[str] = None,
    pattern_split: Optional[Pattern] = None,
    str_prefix: str = "",
) -> List:
    """Turn an ndarray (cell) to a list of strings.

    Parameters
    ----------
    m_cell: np.ndarray
    empty_value: str, optional
        What value to replace empty cells with. Default None.
    pattern_split: Pattern, optional
        Regular expression to use to split the expression. Used for annotations.
        Default None.
    str_prefix: str, optional
        A prefix that will be added to each value in the list if present. Default "".

    Returns
    -------
    List
        A list of processed strings.
    """
    if str_prefix and pattern_split:
        return [
            [
                str_prefix + str_found if str_prefix not in str_found else str_found
                for str_found in pattern_split.findall(str(each_cell[0][0]))
            ]
            if np.size(each_cell[0])
            else empty_value
            for each_cell in m_cell
        ]
    elif pattern_split:
        return [
            pattern_split.findall(str(each_cell[0][0]))
            if np.size(each_cell[0])
            else empty_value
            for each_cell in m_cell
        ]
    else:
        return [
            str(each_cell[0][0]).strip() if np.size(each_cell[0]) else empty_value
            for each_cell in m_cell
        ]


def _cell_to_float_list(
    m_cell: np.ndarray,
    empty_value: Optional[float] = None,
    inf_value: Optional[float] = None,
) -> List:
    """Turn an ndarray (cell) to a list of floats.

    Parameters
    ----------
    m_cell: np.ndarray
    empty_value: float, optional
        What value to replace empty cells with. Default None.
    inf_value: float, optional
        Replace infinite values with defined inf. Default None (don't replace), will be
        replaced by given value  if given.

    Returns
    -------
    List
        A list of processed floats.
    """

    def fix_inf(val: str, _inf_value: float):
        """Fix inf value, used for rxn.lb and rxn.ub.

        Parameters
        ----------
        val: str
            the str to process into float
        _inf_value: float
            A value to replace infite values with.
        """
        val = float(val)
        if np.isinf(val) and val < 0:
            return -_inf_value
        elif np.isinf(val) and val > 0:
            return _inf_value
        else:
            return val

    if inf_value:
        return [
            fix_inf(x[0], _inf_value=inf_value) if np.size(x[0]) else empty_value
            for x in m_cell
        ]
    else:
        return [float(x[0]) if np.size(x[0]) else empty_value for x in m_cell]


def load_matlab_model(
    infile_path: Union[str, Path, IO],
    variable_name: Optional[str] = None,
    inf: float = np.inf,
) -> Model:
    """Load a cobra model stored as a .mat file.

    Parameters
    ----------
    infile_path : str or Path or filehandle
        File path or descriptor of the .mat file describing the cobra model.
    variable_name : str, optional
        The variable name of the model in the .mat file. If None, then the
        first MATLAB variable which looks like a COBRA model will be used
        (default None).
    inf: float, optional
        The value to use for infinite bounds. Some solvers do not handle
        infinite values so for using those, set this to a high numeric value
        (default `numpy.inf`).

    Returns
    -------
    cobra.Model
        The cobra model as represented in the .mat file.

    Raises
    ------
    ImportError
        If scipy is not found in the Python environment.
    IOError
        If no COBRA model is found in the .mat file.

    """
    if not scipy_io:
        raise ImportError("load_matlab_model() requires scipy.")

    if isinstance(infile_path, str):
        data = scipy_io.loadmat(infile_path)
    elif isinstance(infile_path, Path):
        data = scipy_io.loadmat(infile_path.open("rb"))  # noqa W9018
    else:
        data = scipy_io.loadmat(infile_path)  # noqa W9018
    possible_names = []
    if variable_name is None:
        # skip meta variables
        meta_vars = {"__globals__", "__header__", "__version__"}
        possible_names = sorted(i for i in data if i not in meta_vars)
        if len(possible_names) == 1:
            variable_name = possible_names[0]
    elif variable_name is not None:
        return from_mat_struct(data[variable_name], model_id=variable_name, inf=inf)

    for possible_name in possible_names:
        try:
            return from_mat_struct(data[possible_name], model_id=possible_name, inf=inf)
        except ValueError as e:
            print(f"Some problem with the model, causing error {e}")
            # TODO: use custom cobra exception to handle exception
    # If code here is executed, then no model was found.
    raise IOError(f"No COBRA model found at {infile_path}.")


def save_matlab_model(
    model: Model, file_name: Union[str, Path, IO], varname: Optional[str] = None
) -> None:
    """Save the cobra model as a .mat file.

    This .mat file can be used directly in cobratoolbox.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to represent.
    file_name : str or file-like or Path
        File path or descriptor that the MATLAB representation should be
        written to.
    varname : str, optional
       The name of the variable within the MATLAB workspace. Model ID is
       used if available, else 'exported_model' is used (default None).

    """
    if not scipy_io:
        raise ImportError("save_matlab_model() requires scipy.")

    if varname is None:
        varname = (
            str(model.id)
            if model.id is not None and len(model.id) > 0
            else "exported_model"
        )
    mat = create_mat_dict(model)
    scipy_io.savemat(file_name, {varname: mat}, appendmat=True, oned_as="column")


def mat_parse_annotations(
    target_list: List[Object], mat_struct: np.ndarray, d_replace: str = DICT_MET
) -> None:
    """Process mat structure annotations in place.

    Will process mat structured annotations and add them to a list of new entities
    (metabolites, reactions, genes) in a format based on identifiers.org.

    Parameters
    ----------
    target_list: list[cobra.Object]
        A list of cobra objects, including metabolites, reactions or genes. The
        annotations will be added to these lists.
    mat_struct: np.ndarray
        A darray that includes the data imported from matlab file.
    d_replace: str
        A string that points to the dictionary of converstions between MATLAB and
        providers. Default DICT_MET (for metabolite).
    """
    struct_names = [x.casefold() for x in mat_struct.dtype.names]
    matlab_field_dict = {
        x.casefold(): DICT_REPLACE[d_replace][x] for x in DICT_REPLACE[d_replace].keys()
    }
    caseunfold = {x.casefold(): x for x in mat_struct.dtype.names}
    annotation_matlab = list(set(struct_names).intersection(matlab_field_dict.keys()))
    providers = [matlab_field_dict[x] for x in annotation_matlab]
    annotations = dict.fromkeys(providers, None)
    for name, mat_key in zip(providers, annotation_matlab):
        if mat_key == "rxnReferences".casefold():
            # This only picks up PMID: style references. Sometimes there are other
            # things like PMC or OMIM, but those are ignored for now,
            annotations[name] = _cell_to_str_list(
                mat_struct[caseunfold[mat_key]][0, 0], None, _pubmed_re
            )
        elif mat_key == "rxnECNumbers".casefold():
            # turn EC codes to a list
            annotations[name] = _cell_to_str_list(
                mat_struct[caseunfold[mat_key]][0, 0], None, _ec_re
            )
        elif mat_key == "metCHEBIID".casefold():
            annotations[name] = _cell_to_str_list(
                mat_struct[caseunfold[mat_key]][0, 0], None, _chebi_re, "CHEBI:"
            )
        elif mat_key == "metSBOTerms".casefold() or mat_key == "rxnSBOTerms".casefold():
            annotations[name] = _cell_to_str_list(
                mat_struct[caseunfold[mat_key]][0, 0], None, _sbo_re, "SBO:"
            )
        elif mat_key == "metSLMID".casefold():
            annotations[name] = _cell_to_str_list(
                mat_struct[caseunfold[mat_key]][0, 0], None, _sbo_re, "SLM:"
            )
        else:
            # If it something else, which may have commas, turn it into a list
            annotations[name] = [
                [y.strip() for y in each_cell.split(", ")] if each_cell else None
                for each_cell in _cell_to_str_list(
                    mat_struct[caseunfold[mat_key]][0, 0]
                )
            ]
    for i, obj in enumerate(target_list):
        obj.annotation = {
            prov: annotations[prov][i] for prov in providers if annotations[prov][i]
        }

    # TODO - When cobrapy.notes are revised not be a dictionary (possibly when
    #  annotations are fully SBML compliant, revise this function.


def mat_parse_notes(
    target_list: List[Object],
    mat_struct: np.ndarray,
    d_replace: str = DICT_REACTION_NOTES,
) -> None:
    """Process mat structure notes in place.

    Will process mat structured notes and add them to a list of new entities
    (metabolites, reactions, genes) in a format based on identifiers.org.

    Parameters
    ----------
    target_list: list[cobra.Object]
        A list of cobra objects, including metabolites, reactions or genes. The
        notes will be added to these lists.
    mat_struct: np.ndarray
        A darray that includes the data imported from matlab file.
    d_replace: str
        A string that points to the dictionary of converstions between MATLAB and
        notes. Default DICT_REACTION_NOTES (for reactions).
    """
    struct_names = [x.casefold() for x in mat_struct.dtype.names]
    matlab_field_dict = {
        x.casefold(): DICT_REPLACE[d_replace][x] for x in DICT_REPLACE[d_replace].keys()
    }
    caseunfold = {x.casefold(): x for x in mat_struct.dtype.names}
    annotation_matlab = list(set(struct_names).intersection(matlab_field_dict.keys()))
    note_providers = [matlab_field_dict[x] for x in annotation_matlab]
    notes = dict.fromkeys(note_providers, None)
    for name, mat_key in zip(note_providers, annotation_matlab):
        if mat_key == "rxnReferences".casefold():
            # This only removes PMID: style references. Sometimes there are other
            # things like PMC or OMIM, but those are placed as string in notes.
            _notes = _cell_to_str_list(mat_struct[caseunfold[mat_key]][0, 0])
            notes[name] = [
                _pubmed_re.sub("", x).strip()
                if x and len(_pubmed_re.sub("", x).strip())
                else None
                for x in _notes
            ]
        elif mat_key == "rxnConfidenceScores".casefold():
            notes[name] = [
                str(confidence) if confidence is not None else ""
                for confidence in _cell_to_float_list(
                    mat_struct[caseunfold[mat_key]][0, 0]
                )
            ]
        else:
            # If it something else, which may have commas, turn it into a list
            notes[name] = _cell_to_str_list(mat_struct[caseunfold[mat_key]][0, 0])
        notes[name] = [
            _punctuation_re.sub("", _double_punctuation_re.sub("", x)) if x else None
            for x in notes[name]
        ]
    for i, obj in enumerate(target_list):
        obj.notes = {prov: notes[prov][i] for prov in note_providers if notes[prov][i]}


def annotations_to_mat(
    mat_dict: OrderedDict, annotation_list: List[Dict], d_replace: str = DICT_MET_REV
) -> None:
    """Process mat structure annotations in place.

    Will process mat structured annotations and add them to a list of new entities
    (metabolites, reactions, genes) in a format based on identifiers.org.

    Parameters
    ----------
    mat_dict: OrderedDict
        An ordered dictionary having model attributes as keys and their
        respective values represented as arrays, as the values. Annotations will
        be inserted into this OrderdDict.
    annotation_list: list[Dict]
        A list of cobra annotations, in the form of a dictionary.
    d_replace: str
        A string that points to the dictionary of converstions between MATLAB and
        providers. Default DICT_MET_REV (for metabolite).
    """
    providers_used = set()
    for i in range(len(annotation_list)):
        providers_used.update(annotation_list[i].keys())
    providers_used = providers_used.intersection(DICT_REPLACE[d_replace].keys())
    providers_used = list(providers_used)
    annotation_matlab = {prov: DICT_REPLACE[d_replace][prov] for prov in providers_used}
    empty_lists = [[""] * len(annotation_list) for _ in annotation_matlab]
    annotation_cells_to_be = dict(zip(annotation_matlab.values(), empty_lists))
    for i in range(len(annotation_list)):
        for provider_key, obj_annotation in annotation_list[i].items():
            if isinstance(obj_annotation, str):
                obj_annotation = [obj_annotation]
            if provider_key == "pubmed":
                obj_annotation = ", ".join(
                    [
                        "PMID:" + annot if "PMID:" not in annot else annot
                        for annot in obj_annotation
                    ]
                )
            elif provider_key == "CHEBI":
                obj_annotation = ", ".join(
                    [
                        "CHEBI:" + annot if "CHBEI:" not in annot else annot
                        for annot in obj_annotation
                    ]
                )
            elif provider_key == "ec-code":
                obj_annotation = " or ".join(obj_annotation)
            else:
                obj_annotation = ", ".join(obj_annotation)
            if provider_key not in providers_used:
                continue
            annotation_cells_to_be[annotation_matlab[provider_key]][i] = obj_annotation
    for annotation_key, item_list in annotation_cells_to_be.items():
        mat_dict[annotation_key] = _cell(item_list)

    # TODO - When cobrapy.notes are revised not be a dictionary (possibly when
    #  annotations are fully SBML compliant, revise this function.


def notes_to_mat(
    mat_dict: OrderedDict, note_list: List[Dict], d_replace: str = DICT_MET_REV
) -> None:
    """Process mat structure notes in place.

    Will process mat structured annotations and add them to a list of new entities
    (metabolites, reactions, genes) in a format based on identifiers.org.

    Parameters
    ----------
    mat_dict: OrderedDict
        An ordered dictionary having model attributes as keys and their
        respective values represented as arrays, as the values. Annotations will
        be inserted into this OrderdDict.
    note_list: list[Dict]
        A list of cobra annotations, in the form of a dictionary.
    d_replace: str
        A string that points to the dictionary of converstions between MATLAB and
        providers. Default DICT_MET_REV (for metabolite).
    """
    providers_used = set()
    for i in range(len(note_list)):
        providers_used.update(note_list[i].keys())
    providers_used = providers_used.intersection(DICT_REPLACE[d_replace].keys())
    providers_used = list(providers_used)
    annotation_matlab = {prov: DICT_REPLACE[d_replace][prov] for prov in providers_used}
    empty_lists = [[""] * len(note_list) for _ in annotation_matlab]
    annotation_cells_to_be = dict(zip(annotation_matlab.values(), empty_lists))
    for i in range(len(note_list)):
        for provider_key, object_note in note_list[i].items():
            if provider_key not in providers_used:
                continue
            if provider_key == CONFIDENCE_STR:
                object_note = float(object_note)
            if not len(annotation_cells_to_be[annotation_matlab[provider_key]][i]):
                annotation_cells_to_be[annotation_matlab[provider_key]][i] = object_note
            else:
                # References that aren't MIRIAM compliant will go to rxnNotes
                annotation_cells_to_be[annotation_matlab[provider_key]][i] = (
                    annotation_cells_to_be[annotation_matlab[provider_key]][i]
                    + f"; {object_note}"
                )
    for annotation_key, item_list in annotation_cells_to_be.items():
        mat_dict[annotation_key] = _cell(item_list)


def create_mat_dict(model: Model) -> OrderedDict:
    """Create a dictionary mapping model attributes to arrays.

    Parameters
    ----------
    model : cobra.Model
        The model to create dictionary for.

    Returns
    -------
    OrderedDict
        The ordered dictionary having model attributes as keys and their
        respective values represented as arrays, as the values.

    """
    rxns = model.reactions
    mets = model.metabolites
    mat = OrderedDict()
    mat["mets"] = _cell(mets.list_attr("id"))
    model_has_compartment_names = False
    if list(model.compartments.keys()) != list(model.compartments.values()):
        model_has_compartment_names = True
    if {_get_id_compartment(met_id) for met_id in mets.list_attr("id")} == {
        None
    } or model_has_compartment_names:
        comps = list(model.compartments.keys())
        mat["comps"] = _cell(comps)
        mat["compNames"] = _cell([model.compartments[comp] for comp in comps])
        mat["metComps"] = [comps.index(x) + 1 for x in mets.list_attr("compartment")]
    mat["metNames"] = _cell(mets.list_attr("name"))
    mat["metFormulas"] = _cell([str(m.formula) if m.formula else "" for m in mets])
    try:
        mat["metCharges"] = np.array(mets.list_attr("charge")) * 1.0
    except (TypeError, AttributeError):
        # can't have any None entries for charge, or this will fail
        # TODO: use custom cobra exception to handle exception
        pass
    annotations_to_mat(mat, mets.list_attr("annotation"), DICT_MET_REV)
    # TODO - When cobrapy.notes are revised not be a dictionary (possibly when
    #  annotations are fully SBML compliant, revise this function.
    notes_to_mat(mat, mets.list_attr("notes"), DICT_MET_NOTES_REV)
    mat["genes"] = _cell(model.genes.list_attr("id"))
    gene_names = model.genes.list_attr("name")
    if not all(_name == "" for _name in gene_names):
        mat["geneNames"] = _cell(gene_names)
    annotations_to_mat(mat, model.genes.list_attr("annotation"), DICT_GENE_REV)
    # make a matrix for rxnGeneMat
    # reactions are rows, genes are columns
    rxn_gene = scipy_sparse.dok_matrix((len(model.reactions), len(model.genes)))
    if min(rxn_gene.shape) > 0:
        for i, reaction in enumerate(model.reactions):
            for gene in reaction.genes:
                rxn_gene[i, model.genes.index(gene)] = 1
        mat["rxnGeneMat"] = rxn_gene
    mat["grRules"] = _cell(rxns.list_attr("gene_reaction_rule"))
    mat["rxns"] = _cell(rxns.list_attr("id"))
    mat["rxnNames"] = _cell(rxns.list_attr("name"))
    annotations_to_mat(mat, rxns.list_attr("annotation"), DICT_REACTION_REV)
    notes_to_mat(mat, rxns.list_attr("notes"), DICT_REACTION_NOTES_REV)
    rxn_subsystems = _cell(rxns.list_attr("subsystem"))
    group_ids = model.groups.list_attr("id")
    group_names = model.groups.list_attr("name")
    group_ids = [
        _id if not _name else _name for _id, _name in zip(group_ids, group_names)
    ]
    group_members_list = model.groups.list_attr("members")
    if set(rxn_subsystems) == {""} and set(group_ids) != {""}:
        subsystems = [[] for _ in rxns]
        for group_id, group_members in zip(group_ids, group_members_list):
            group = model.groups.get_by_id(group_id)
            if group.kind == "partonomy":
                for member in group_members:
                    if isinstance(member, Reaction):
                        rxn_ind = model.reactions.index(member)
                        # noinspection PyTypeChecker
                        subsystems[rxn_ind].append(group_id)
        mat["subSystems"] = _cell(
            [", ".join(subsystem_list) for subsystem_list in subsystems]
        )
    else:
        mat["subSystems"] = _cell(rxns.list_attr("subsystem"))
    stoich_mat = create_stoichiometric_matrix(model, array_type="dok")
    mat["S"] = stoich_mat if stoich_mat is not None else [[]]
    # multiply by 1 to convert to float, working around scipy bug
    # https://github.com/scipy/scipy/issues/4537
    mat["lb"] = np.array(rxns.list_attr("lower_bound")) * 1.0
    mat["ub"] = np.array(rxns.list_attr("upper_bound")) * 1.0
    mat["b"] = np.array(mets.list_attr("_bound")) * 1.0
    mat["c"] = np.array(rxns.list_attr("objective_coefficient")) * 1.0
    mat["rev"] = np.array(rxns.list_attr("reversibility")) * 1
    if model.name:
        mat["modelName"] = str(model.name)
    mat["description"] = str(model.id)
    return mat


def from_mat_struct(
    mat_struct: np.ndarray,
    model_id: Optional[str] = None,
    inf: float = np.inf,
) -> Model:
    """Create a model from the cobratoolbox struct.

    Parameters
    ----------
    mat_struct : numpy.ndarray
        The `numpy.ndarray` that most likely contains the model, being chosen by
        load_matlab_file after loading the matlab structure via scipy_io.loadmat.
    model_id : str, optional
        The ID of the model generated. If None, will try to look for ID in
        model's description. If multiple IDs are found, the first one is
        used. If no IDs are found, will use 'imported_model' (default None).
    inf : float, optional
        The value to use for infinite bounds. Some solvers do not handle
        infinite values so for using those, set this to a high numeric value
        (default `numpy.inf`).

    Returns
    -------
    cobra.Model
        The model as represented in .mat file.

    """
    m = mat_struct

    if m.dtype.names is None or {"rxns", "mets", "S", "lb", "ub"} > set(m.dtype.names):
        raise ValueError("Invalid MATLAB struct.")

    old_cobratoolbox_fields = [
        "confidenceScores",
        "metCharge",
        "ecNumbers",
        "KEGGID",
        "metSmile",
        "metHMDB",
    ]
    new_cobratoolbox_fields = [
        "rxnConfidenceScores",
        "metCharges",
        "rxnECNumbers",
        "metKEGGID",
        "metSmiles",
        "metHMDBID",
    ]
    for old_field, new_field in zip(old_cobratoolbox_fields, new_cobratoolbox_fields):
        if old_field in m.dtype.names and new_field not in m.dtype.names:
            logger.warning(
                f"This model seems to have {old_field} instead of {new_field} field. "
                f"Will use {old_field} for what {new_field} represents."
            )
            new_names = list(m.dtype.names)
            new_names[new_names.index(old_field)] = new_field
            m.dtype.names = new_names

    model = Model()
    if model_id is not None:
        model.id = model_id
    elif "description" in m.dtype.names:
        description = m["description"][0, 0][0]
        if not isinstance(description, str) and len(description) > 1:
            model.id = description[0]
            logger.warning("Several IDs detected, only using the first.")
        else:
            model.id = description
    else:
        model.id = "imported_model"
    if "modelName" in m.dtype.names and np.size(m["modelName"][0, 0]):
        model.name = m["modelName"][0, 0][0]

    met_ids = _cell_to_str_list(m["mets"][0, 0])
    if {"metComps", "comps", "compNames"}.issubset(m.dtype.names):
        met_comp_index = [x[0] - 1 for x in m["metComps"][0][0]]
        comps = _cell_to_str_list(m["comps"][0, 0])
        comp_names = _cell_to_str_list(m["compNames"][0][0])
        met_comps = [comps[i] for i in met_comp_index]
        met_comp_names = [comp_names[i] for i in met_comp_index]
    else:
        logger.warning(
            f"No defined compartments in model {model.id}. "
            f"Compartments will be deduced heuristically "
            f"using regular expressions."
        )
        met_comps = [_get_id_compartment(x) for x in met_ids]
        met_comp_names = met_comps
        if None in met_comps or "" in met_comps:
            raise ValueError("Some compartments were empty. Check the model!")
        logger.warning(
            f"Using regular expression found the following compartments:"
            f"{', '.join(sorted(set(met_comps)))}"
        )
    if None in met_comps or "" in met_comps:
        raise ValueError("Some compartments were empty. Check the model!")
    model.compartments = dict(zip(met_comps, met_comp_names))
    met_names, met_formulas, met_charges = None, None, None
    try:
        met_names = _cell_to_str_list(m["metNames"][0, 0], "")
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    try:
        met_formulas = _cell_to_str_list(m["metFormulas"][0, 0])
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    try:
        met_charges = _cell_to_float_list(m["metCharges"][0, 0])
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    new_metabolites = []
    for i in range(len(met_ids)):
        new_metabolite = Metabolite(met_ids[i], compartment=met_comps[i])
        if met_names:
            new_metabolite.name = met_names[i]
        if met_charges:
            new_metabolite.charge = met_charges[i]
        if met_formulas:
            new_metabolite.formula = met_formulas[i]
        new_metabolites.append(new_metabolite)
    mat_parse_annotations(new_metabolites, m, d_replace=DICT_MET)
    mat_parse_notes(new_metabolites, m, d_replace=DICT_MET_NOTES)
    model.add_metabolites(new_metabolites)

    if "genes" in m.dtype.names:
        gene_names = None
        gene_ids = _cell_to_str_list(m["genes"][0, 0])
        try:
            gene_names = _cell_to_str_list(m["geneNames"][0, 0])
        except (IndexError, ValueError):
            # TODO: use custom cobra exception to handle exception
            pass
        new_genes = [
            Gene(gene_ids[i], name=gene_names[i]) if gene_names else Gene(gene_ids[i])
            for i in range(len(gene_ids))
        ]
        mat_parse_annotations(new_genes, m, d_replace=DICT_GENE)
        for current_gene in new_genes:
            current_gene._model = model
        model.genes += new_genes

    new_reactions = []
    rxn_ids = _cell_to_str_list(m["rxns"][0, 0])
    rxn_lbs = _cell_to_float_list(m["lb"][0, 0], empty_value=None, inf_value=inf)
    rxn_ubs = _cell_to_float_list(m["ub"][0, 0], empty_value=None, inf_value=inf)
    rxn_gene_rules, rxn_names, rxn_subsystems = None, None, None
    try:
        rxn_gene_rules = _cell_to_str_list(m["grRules"][0, 0], "")
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    try:
        rxn_names = _cell_to_str_list(m["rxnNames"][0, 0], "")
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    try:
        # RECON3.0 mat has an array within an array for subsystems.
        # If we find a model that has multiple subsytems per reaction, this should be
        # modified
        if np.sctype2char(m["subSystems"][0, 0][0][0]) == "O" and isinstance(
            m["subSystems"][0, 0][0][0][0], np.ndarray
        ):
            rxn_subsystems = [
                each_cell[0][0][0][0] if each_cell else ""
                for each_cell in m["subSystems"][0, 0]
            ]
        # Other matlab files seem normal.
        else:
            rxn_subsystems = _cell_to_str_list(m["subSystems"][0, 0], "")
    except (IndexError, ValueError):
        # TODO: use custom cobra exception to handle exception
        pass
    for i in range(len(rxn_ids)):
        new_reaction = Reaction(
            id=rxn_ids[i], lower_bound=rxn_lbs[i], upper_bound=rxn_ubs[i]
        )
        if rxn_names:
            new_reaction.name = rxn_names[i]
        if rxn_subsystems:
            new_reaction.subsystem = rxn_subsystems[i]
        if rxn_gene_rules:
            new_reaction.gene_reaction_rule = rxn_gene_rules[i]
        new_reactions.append(new_reaction)
    mat_parse_annotations(new_reactions, m, d_replace=DICT_REACTION)
    # TODO - When cobrapy.notes are revised not be a dictionary (possibly when
    #  annotations are fully SBML compliant, revise this function.
    mat_parse_notes(new_reactions, m, d_replace=DICT_REACTION_NOTES)

    csc = scipy_sparse.csc_matrix(m["S"][0, 0])
    for i in range(csc.shape[1]):
        stoic_dict = {
            model.metabolites[j]: csc[j, i] for j in csc.getcol(i).nonzero()[0]
        }
        new_reactions[i].add_metabolites(stoic_dict)

    model.add_reactions(new_reactions)

    # Make subsystems into groups
    if rxn_subsystems:
        rxn_group_names = set(rxn_subsystems).difference({None})
        new_groups = []
        for g_name in sorted(rxn_group_names):
            group_members = model.reactions.query(
                lambda x: x.subsystem == g_name  # noqa: B023
            )
            new_group = Group(
                id=g_name, name=g_name, members=group_members, kind="partonomy"
            )
            new_group.annotation["sbo"] = "SBO:0000633"
            new_groups.append(new_group)
        model.add_groups(new_groups)

    if "c" in m.dtype.names:
        c_vec = _cell_to_float_list(m["c"][0, 0])
        coefficients = dict(zip(new_reactions, c_vec))
        set_objective(model, coefficients)
    else:
        logger.warning("Objective vector `c` not found.")

    if "osenseStr" in m.dtype.names:
        if isinstance(m["osenseStr"][0, 0][0], np.str_):
            model.objective_direction = str(m["osenseStr"][0, 0][0])
        elif isinstance(m["osenseStr"][0, 0][0], np.ndarray):
            model.objective_direction = str(m["osenseStr"][0, 0][0][0])
    elif "osense" in m.dtype.names:
        osense = float(m["osense"][0, 0][0][0])
        objective_direction_str = "max"
        if osense == 1:
            objective_direction_str = "min"
        model.objective_direction = objective_direction_str
    return model



"""Provide functions for loading and saving metabolic models."""


from cobra.io.dict import model_from_dict, model_to_dict
from cobra.io.json import from_json, load_json_model, save_json_model, to_json
from cobra.io.mat import load_matlab_model, save_matlab_model
from cobra.io.sbml import read_sbml_model, write_sbml_model, validate_sbml_model
from cobra.io.yaml import from_yaml, load_yaml_model, save_yaml_model, to_yaml
from cobra.io.web import AbstractModelRepository, BiGGModels, BioModels, load_model



"""Provide functions for I/O in JSON format."""
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Union

from .dict import model_from_dict, model_to_dict


try:
    import simplejson as json
except ImportError:
    import json

if TYPE_CHECKING:
    from cobra import Model


JSON_SPEC = "1"


def to_json(model: "Model", sort: bool = False, **kwargs: Any) -> str:
    """Return the model as a JSON string.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to represent.
    sort : bool, optional
        Whether to sort the metabolites, reactions, and genes or maintain the
        order defined in the model (default False).
    **kwargs : Any
        Keyword arguments passed on to `json.dumps`.

    Returns
    -------
    str
        JSON string representation of the cobra model.

    See Also
    --------
    save_json_model : Write directly to a file.
    json.dumps : Base function.

    """
    obj = model_to_dict(model, sort=sort)
    obj["version"] = JSON_SPEC
    return json.dumps(obj, allow_nan=False, **kwargs)


def from_json(document: str) -> "Model":
    """Load a cobra model from a JSON string.

    Parameters
    ----------
    document : str
        The JSON string representation of a cobra model.

    Returns
    -------
    cobra.Model
        The cobra model as interpreted from the JSON string.

    See Also
    --------
    load_json_model : Load directly from a file.

    """
    return model_from_dict(json.loads(document))


def save_json_model(
    model: "Model",
    filename: Union[str, Path, IO],
    sort: bool = False,
    pretty: bool = False,
    **kwargs: Any
) -> None:
    """Write the cobra model to a file in JSON format.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to represent.
    filename : str or file-like
        File path or descriptor that the JSON representation should be
        written to.
    sort : bool, optional
        Whether to sort the metabolites, reactions, and genes or maintain the
        order defined in the model (default False).
    pretty : bool, optional
        Whether to format the JSON more compactly (default) or in a more
        verbose but easier to read fashion. Can be partially overwritten by the
        `**kwargs` (default False).
    **kwargs : Any
        Keyword arguments passed to `json.dump`.

    See Also
    --------
    to_json : Return a string representation.
    json.dump : Base function.

    """
    obj = model_to_dict(model, sort=sort)
    obj["version"] = JSON_SPEC

    if pretty:
        dump_opts = {
            "indent": 4,
            "separators": (",", ": "),
            "sort_keys": True,
            "allow_nan": False,
        }
    else:
        dump_opts = {
            "indent": 0,
            "separators": (",", ":"),
            "sort_keys": False,
            "allow_nan": False,
        }
    dump_opts.update(**kwargs)

    if isinstance(filename, (str, Path)):
        with open(filename, "w") as file_handle:
            json.dump(obj, file_handle, **dump_opts)
    else:
        json.dump(obj, filename, **dump_opts)


def load_json_model(filename: Union[str, Path, IO]) -> "Model":
    """Load a cobra model from a file in JSON format.

    Parameters
    ----------
    filename : str or file-like
        File path or descriptor that contains the JSON document describing the
        cobra model.

    Returns
    -------
    cobra.Model
        The cobra model as represented in the JSON document.

    See Also
    --------
    from_json : Load from a JSON string.

    """
    if isinstance(filename, (str, Path)):
        with open(filename, "r") as file_handle:
            return model_from_dict(json.load(file_handle))
    else:
        return model_from_dict(json.load(filename))



"""Provide the abstract base summary class."""


import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

import pandas as pd


if TYPE_CHECKING:
    from cobra import Model, Solution


logger = logging.getLogger(__name__)


class Summary(ABC):
    """
    Define the abstract base summary.

    See Also
    --------
    MetaboliteSummary
    ReactionSummary
    ModelSummary

    """

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        Initialize a summary.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        """
        super().__init__(**kwargs)
        self._flux = None
        self._tolerance = None

    def _generate(
        self,
        model: "Model",
        solution: Optional["Solution"],
        fva: Optional[Union[float, pd.DataFrame]],
    ) -> None:
        """
        Prepare the data for the summary instance.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model for which to generate a metabolite summary.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution.
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched.

        """
        self._tolerance = model.tolerance

    def __str__(self) -> str:
        """Return a string representation of the summary."""
        return self.to_string()

    def _repr_html_(self) -> str:
        """Return a rich HTML representation of the summary."""
        return self.to_html()

    @property
    def tolerance(self) -> float:
        """Return the set threshold."""
        return self._tolerance

    def _normalize_threshold(self, threshold: Optional[float]):
        """Return a sensible threshold value."""
        if threshold is None:
            return self.tolerance
        else:
            if threshold < self.tolerance:
                logger.warning(
                    "The chosen threshold cannot be less than the model's "
                    "tolerance value. Using the tolerance instead."
                )
                return self.tolerance
            else:
                return threshold

    @abstractmethod
    def to_string(
        self,
        names: bool = False,
        threshold: Optional[float] = None,
        float_format: str = ".4G",
        column_width: int = 79,
    ) -> str:
        """
        Return a pretty string representation of the summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').
        column_width : int, optional
            The maximum column width for each row (default 79).

        Returns
        -------
        str
            The summary formatted as a pretty string.

        """
        raise NotImplementedError(
            "This method needs to be implemented by the subclass."
        )

    @abstractmethod
    def to_html(
        self,
        names: bool = False,
        threshold: Optional[float] = None,
        float_format: str = ".4G",
    ) -> str:
        """
        Return a rich HTML representation of the metabolite summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').

        Returns
        -------
        str
            The summary formatted as HTML.

        """
        raise NotImplementedError(
            "This method needs to be implemented by the subclass."
        )

    def to_frame(self) -> pd.DataFrame:
        """Return the a data frame representation of the summary."""
        return self._flux.copy()



"""Define base Object class in Cobra."""

from typing import Optional


class Object:
    """Defines common behavior of object in cobra.core."""

    def __init__(self, id: Optional[str] = None, name: str = "") -> None:
        """Initialize a simple object with an identifier.

        Parameters
        ----------
        id: string, optional
            the identifier to associate with the object
        name: string, optional
            The name to associate with the object. Default "".

        Objects will have notes and _annotation as dicitionaries, initialized as empty
        dictionaries.
        """
        self._id = id
        self.name = name

        self.notes = {}
        self._annotation = {}

    @property
    def id(self) -> str:
        """Get the Object id.

        Returns
        -------
        id: str
        """
        return getattr(self, "_id", None)

    @id.setter
    def id(self, value) -> None:
        """Set the id to value.

        Parameters
        ----------
        value: str
            The string to set the id to.

        Raises
        ------
        TypeError if value is not a string.
        """
        if value == self.id:
            pass
        elif not isinstance(value, str):
            raise TypeError("ID must be a string")
        elif getattr(self, "_model", None) is not None:
            self._set_id_with_model(value)
        else:
            self._id = value

    def _set_id_with_model(self, value) -> None:
        """Set id with model.

        This appears to be a stub so it can be modified in dependant classes.

        Parameters
        ----------
        value: str
            The string to set the id to.
        """
        self._id = value

    @property
    def annotation(self) -> dict:
        """Get annotation dictionary.

        Returns
        -------
        _annotation: dict
            Returns _annotation as a dictionary.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, annotation):
        """Set annotation.

        Parameters
        ----------
        annotation: dict
            Annotation dictionary to set _annotation to. Will raise error if not dict.

        Raises
        ------
        TypeError if annotation not a dict.
        """
        if not isinstance(annotation, dict):
            raise TypeError("Annotation must be a dict")
        else:
            self._annotation = annotation

    def __getstate__(self) -> dict:
        """Get state of annotation.

        To prevent excessive replication during deepcopy, ignores _model in state.

        Returns
        -------
        state: dict
            Dictionary of state, excluding _model.
        """
        state = self.__dict__.copy()
        if "_model" in state:
            state["_model"] = None
        return state

    def __repr__(self) -> str:
        """Return string representation of Object, with class.

        Returns
        -------
        str
            Composed of class.name, id and hexadecimal of id.
        """
        return f"<{self.__class__.__name__} {self.id} at {id(self):#x}>"

    def __str__(self) -> str:
        """Return string representation of object.

        Returns
        -------
        str
            Object.id as string.
        """
        return str(self.id)



"""Provide a process pool with enhanced performance on Windows."""


import multiprocessing
import os
import pickle
from pathlib import Path
from platform import system
from tempfile import mkstemp
from types import TracebackType
from typing import Any, Callable, Optional, Tuple, Type


__all__ = ("ProcessPool",)


def _init_win_worker(filename: str) -> None:
    """Retrieve worker initialization code from a pickle file and call it."""
    with open(filename, mode="rb") as handle:
        func, *args = pickle.load(handle)
    func(*args)


class ProcessPool:
    """Define a process pool that handles the Windows platform specially."""

    def __init__(
        self,
        processes: Optional[int] = None,
        initializer: Optional[Callable] = None,
        initargs: Tuple = (),
        maxtasksperchild: Optional[int] = None,
        **kwargs
    ) -> None:
        """
        Initialize a process pool.

        Add a thin layer on top of the `multiprocessing.Pool` that, on Windows, passes
        initialization code to workers via a pickle file rather than directly. This is
        done to avoid a performance issue that exists on Windows. Please, also see the
        discussion [1_].

        References
        ----------
        .. [1] https://github.com/opencobra/cobrapy/issues/997

        """
        super().__init__(**kwargs)
        self._filename = None
        if initializer is not None and system() == "Windows":
            descriptor, self._filename = mkstemp(suffix=".pkl")
            # We use the file descriptor to the open file returned by `mkstemp` to
            # ensure that the resource is closed and can later be removed. Otherwise
            # Windows will cause a `PermissionError`.
            with os.fdopen(descriptor, mode="wb") as handle:
                pickle.dump((initializer,) + initargs, handle)
            initializer = _init_win_worker
            initargs = (self._filename,)
        self._pool = multiprocessing.Pool(
            processes=processes,
            initializer=initializer,
            initargs=initargs,
            maxtasksperchild=maxtasksperchild,
        )

    def __getattr__(self, name: str, **kwargs) -> Any:
        """Defer attribute access to the pool instance."""
        return getattr(self._pool, name, **kwargs)

    def __enter__(self) -> "ProcessPool":
        """Enable context management."""
        self._pool.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        """Clean up resources when leaving a context."""
        # The `multiprocessing.Pool.__exit__` only terminates pool processes. For a
        # clean exit, we close the pool and join the pool processes first.
        try:
            self._pool.close()
            self._pool.join()
        finally:
            self._clean_up()
        result = self._pool.__exit__(exc_type, exc_val, exc_tb)
        return result

    def close(self) -> None:
        """
        Close the process pool.

        Prevent any more tasks from being submitted to the pool. Once all the tasks have
        been completed, the worker processes will exit.

        """
        try:
            self._pool.close()
        finally:
            self._clean_up()

    def _clean_up(self) -> None:
        """Remove the dump file if it exists."""
        if self._filename is not None and Path(self._filename).exists():
            Path(self._filename).unlink()



"""Provide functionality to access remote model repositories."""


from .abstract_model_repository import AbstractModelRepository
from .bigg_models_repository import BiGGModels
from .biomodels_repository import BioModels
from .load import load_model



"""Provide a global configuration object."""


import logging
import pathlib
import types
from numbers import Number
from os import cpu_count
from textwrap import dedent
from typing import Optional, Tuple, Union

import appdirs

from ..exceptions import SolverNotFound
from ..util.solver import interface_to_str
from ..util.solver import solvers as SOLVERS
from .singleton import Singleton


__all__ = ("Configuration",)


logger = logging.getLogger(__name__)


class Configuration(metaclass=Singleton):
    """
    Define a global configuration object.

    The attributes of this singleton object are used as default values by cobra
    functions.

    Attributes
    ----------
    solver : {"glpk", "cplex", "gurobi", "glpk_exact"}
        The default solver for new models. The solver choices are the ones
        provided by `optlang` and depend on solvers installed in your environment.
    tolerance : float, optional
        The default tolerance for the solver being used (default 1E-07).
    lower_bound : float, optional
        The standard lower bound for reversible reactions (default -1000).
    upper_bound : float, optional
        The standard upper bound for all reactions (default 1000).
    bounds : tuple of floats
        The default reaction bounds for newly created reactions. The bounds
        are in the form of lower_bound, upper_bound (default -1000.0, 1000.0).
    processes : int > 0
        A default number of processes to use where multiprocessing is
        possible. The default number corresponds to the number of available
        cores (hyperthreads) minus one.
    cache_directory : pathlib.Path or str, optional
        A path where the model cache should reside if caching is desired. The
        default directory depends on the operating system.
    max_cache_size : int, optional
        The allowed maximum size of the model cache in bytes (default 1 GB).
    cache_expiration : int, optional
        The expiration time in seconds for the model cache if any (default None).

    """

    def __init__(self, **kwargs) -> None:
        """Initialize the configuration with its default attribute values."""
        super().__init__(**kwargs)
        self._solver = None
        self.tolerance = 1e-07
        self.lower_bound = None
        self.upper_bound = None
        self.processes = None
        self._cache_directory = None
        # Set the cache size to a maximum of 100 MB.
        self.max_cache_size = 100 * (1024**2)
        self.cache_expiration = None

        self.bounds = -1000.0, 1000.0
        self._set_default_solver()
        self._set_default_processes()
        self._set_default_cache_directory()

    def _set_default_solver(self) -> None:
        """Set the default solver from a preferred order."""
        for name in ["gurobi", "cplex", "glpk"]:
            try:
                self.solver = name
            except SolverNotFound:
                continue
            else:
                break

    def _set_default_processes(self) -> None:
        """Set the default number of processes."""
        self.processes = cpu_count()
        if self.processes is None:
            logger.warning("The number of cores could not be detected - assuming one.")
            self.processes = 1
        if self.processes > 1:
            self.processes -= 1

    def _set_default_cache_directory(self) -> None:
        """Set the platform-dependent default cache directory."""
        self.cache_directory = pathlib.Path(
            appdirs.user_cache_dir(appname="cobrapy", appauthor="opencobra")
        )

    @property
    def solver(self) -> types.ModuleType:
        """Return the optlang solver interface."""
        return self._solver

    @solver.setter
    def solver(self, value) -> None:
        """Set the optlang solver interface."""
        not_valid_interface = SolverNotFound(
            f"'{value}' is not a valid solver interface. "
            f" Please pick one from {', '.join(SOLVERS)}."
        )
        if isinstance(value, str):
            if value not in SOLVERS:
                raise not_valid_interface
            interface = SOLVERS[value]
        elif isinstance(value, types.ModuleType) and hasattr(value, "Model"):
            interface = value
        else:
            raise not_valid_interface
        self._solver = interface

    @property
    def bounds(self) -> Tuple[Optional[Number], Optional[Number]]:
        """Return the lower, upper reaction bound pair.

        Returns
        -------
        tuple of number and number or None and None
            The lower and upper bounds for new reactions.

        """
        return self.lower_bound, self.upper_bound

    @bounds.setter
    def bounds(self, bounds: Tuple[Optional[Number], Optional[Number]]) -> None:
        """Set the lower, upper reaction bound pair.

        Parameters
        ----------
        bounds : tuple of number and number or None and None
            The lower and upper bounds for new reactions.

        """
        if None not in bounds:
            assert bounds[0] <= bounds[1]
        self.lower_bound = bounds[0]
        self.upper_bound = bounds[1]

    @property
    def cache_directory(self) -> pathlib.Path:
        """Return the model cache directory."""
        return self._cache_directory

    @cache_directory.setter
    def cache_directory(self, path: Union[pathlib.Path, str]) -> None:
        """
        Set the model cache directory.

        The directory path is created if it doesn't exist yet.

        Parameters
        ----------
        path : pathlib.Path or str
            The path to the cache directory.

        """
        self._cache_directory = pathlib.Path(path)
        if not self._cache_directory.is_dir():
            logger.debug(f"Creating cache directory '{str(self._cache_directory)}'.")
            self._cache_directory.mkdir(parents=True)

    def __repr__(self) -> str:
        """Return a string representation of the current configuration values."""
        return dedent(
            f"""
            solver: {interface_to_str(self.solver)}
            tolerance: {self.tolerance}
            lower_bound: {self.lower_bound}
            upper_bound: {self.upper_bound}
            processes: {self.processes}
            cache_directory: {self.cache_directory}
            max_cache_size: {self.max_cache_size}
            cache_expiration: {self.cache_expiration}
            """
        )

    def _repr_html_(self) -> str:
        """
        Return a rich HTML representation of the current configuration values.

        Notes
        -----
        This special method is used automatically in Jupyter notebooks to display a
        result from a cell.

        """
        return dedent(
            f"""
            <table>
              <thead>
                <tr>
                  <td><strong>Attribute</strong></td>
                  <td><strong>Description</strong></td>
                  <td><strong>Value</strong></td>
                </tr>
              </thead>
              <tbody>
                <tr>
                  <td><pre>solver</pre></td>
                  <td>Mathematical optimization solver</td>
                  <td>{interface_to_str(self.solver)}</td>
                </tr>
                <tr>
                    <td><pre>tolerance</pre></td>
                    <td>General solver tolerance (feasibility, integrality, etc.)</td>
                    <td>{self.tolerance}</td>
                </tr>
                <tr>
                    <td><pre>lower_bound</pre></td>
                    <td>Default reaction lower bound</td>
                    <td>{self.lower_bound}</td>
                </tr>
                <tr>
                    <td><pre>upper_bound</pre></td>
                    <td>Default reaction upper bound</td>
                    <td>{self.upper_bound}</td>
                </tr>
                <tr>
                    <td><pre>processes</pre></td>
                    <td>Number of parallel processes</td>
                    <td>{self.processes}</td>
                </tr>
                <tr>
                    <td><pre>cache_directory</pre></td>
                    <td>Path for the model cache</td>
                    <td>{self.cache_directory}</td>
                </tr>
                <tr>
                    <td><pre>max_cache_size</pre></td>
                    <td>Maximum cache size in bytes</td>
                    <td>{self.max_cache_size}</td>
                </tr>
                <tr>
                    <td><pre>cache_expiration</pre></td>
                    <td>Model cache expiration time in seconds (if any)</td>
                    <td>{self.cache_expiration}</td>
                </tr>
              </tbody>
            </table>
            """
        )



"""Helper functions for all flux analysis methods."""

from typing import TYPE_CHECKING, Optional


if TYPE_CHECKING:
    from cobra import Model


def normalize_cutoff(model: "Model", zero_cutoff: Optional[float] = None) -> float:
    """Return a valid zero cutoff value.

    Parameters
    ----------
    model : cobra.Model
        The model to operate on.
    zero_cutoff : positive float, optional
        The zero cutoff value. If not specified, defaults to
        `model.tolerance` (default None).

    Returns
    -------
    float
        The normalized zero cutoff value.

    Raises
    ------
    ValueError
        If the specified `zero_cutoff` is lesser than `model.tolerance`.

    """
    if zero_cutoff is None:
        return model.tolerance
    else:
        if zero_cutoff < model.tolerance:
            raise ValueError(
                "The chosen zero cutoff cannot be less than the model's "
                "tolerance value."
            )
        else:
            return zero_cutoff



__author__ = "The cobrapy core development team."
__version__ = "0.26.3"


from cobra.core import (
    Configuration,
    DictList,
    Gene,
    Metabolite,
    Model,
    Object,
    Reaction,
    Solution,
    Species,
)
from cobra import flux_analysis
from cobra import io
from cobra import medium
from cobra import sampling
from cobra import summary
from cobra.util import show_versions



"""Provide functions for model component validations."""

from typing import TYPE_CHECKING, Dict, List


if TYPE_CHECKING:
    from cobra import Metabolite, Model, Reaction


# Set of mass unbalanced SBO terms
_NOT_MASS_BALANCED_TERMS = {
    "SBO:0000627",  # EXCHANGE
    "SBO:0000628",  # DEMAND
    "SBO:0000629",  # BIOMASS
    "SBO:0000631",  # PSEUDOREACTION
    "SBO:0000632",  # SINK
}


def check_mass_balance(model: "Model") -> Dict["Reaction", Dict["Metabolite", float]]:
    """Check mass balance for reactions of `model` and return unbalanced ones.

    Parameters
    ----------
    model: cobra.Model
        The model to perform check on.

    Returns
    -------
    dict of {cobra.Reaction: dict of {cobra.Metabolite: float}}
        Returns an empty dict if all components are balanced.

    """
    unbalanced = {}
    for reaction in model.reactions:
        if reaction.annotation.get("sbo") not in _NOT_MASS_BALANCED_TERMS:
            balance = reaction.check_mass_balance()
            if balance:
                unbalanced[reaction] = balance
    return unbalanced


def check_metabolite_compartment_formula(model: "Model") -> List[str]:
    """Check metabolite formulae of `model`.

    Parameters
    ----------
    model: cobra.Model
        The model to perform check on.

    Returns
    -------
    list of str
        Returns an empty list if no errors are found.

    """
    errors = []
    for met in model.metabolites:
        if met.formula is not None and len(met.formula) > 0:
            if not met.formula.isalnum():
                errors.append(
                    f"Metabolite '{met.id}' formula '{met.formula}' not alphanumeric"
                )
    return errors



"""Define the Metabolite class."""

import re
from typing import TYPE_CHECKING, Dict, Optional, Union
from warnings import warn

from ..exceptions import OptimizationError
from ..util.solver import check_solver_status
from ..util.util import format_long_string
from .formula import elements_and_molecular_weights
from .species import Species


if TYPE_CHECKING:
    from optlang.interface import Container
    from pandas import DataFrame

    from cobra.core import Solution
    from cobra.summary.metabolite_summary import MetaboliteSummary

# Numbers are not required because of the |(?=[A-Z])? block. See the
# discussion in https://github.com/opencobra/cobrapy/issues/128 for
# more details.
element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")


class Metabolite(Species):
    """Class for information about metabolite in cobra.Reaction.

    Metabolite is a class for holding information regarding
    a metabolite in a cobra.Reaction object.

    Parameters
    ----------
    id : str
        the identifier to associate with the metabolite
    formula : str
        Chemical formula (e.g. H2O)
    name : str
        A human readable name.
    charge : float
       The charge number of the metabolite
    compartment: str or None
       Compartment of the metabolite.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        id: Optional[str] = None,
        formula: Optional[str] = None,
        name: Optional[str] = "",
        charge: Optional[float] = None,
        compartment: Optional[str] = None,
    ) -> None:
        """Initialize Metaboblite cobra Species.

        Parameters
        ----------
        id : str
            the identifier to associate with the metabolite
        formula : str
            Chemical formula (e.g. H2O)
        name : str
            A human readable name.
        charge : float
           The charge number of the metabolite
        compartment: str or None
           Compartment of the metabolite.
        """
        super().__init__(id=id, name=name)
        self.formula = formula
        # because in a Model a metabolite may participate in multiple Reactions
        self.compartment = compartment
        self.charge = charge

        self._bound = 0.0

    def _set_id_with_model(self, value: str) -> None:
        """Set id with value.

        Parameters
        ----------
        value: str
        """
        if value in self.model.metabolites:
            raise ValueError(
                f"The model already contains a metabolite with the id:" f" {value}"
            )
        self.model.constraints[self.id].name = value
        self._id = value
        self.model.metabolites._generate_index()

    @property
    def constraint(self) -> "Container":
        """Get the constraints associated with this metabolite from the solver.

        Returns
        -------
        optlang.<interface>.Containter
            the optlang constraint for this metabolite
        """
        if self.model is not None:
            return self.model.constraints[self.id]

    @property
    def elements(self) -> Optional[Dict[str, Union[int, float]]]:
        """Get dicitonary of elements and counts.

        Dictionary of elements as keys and their count in the metabolite
        as integer. When set, the `formula` property is updated accordingly.

        Returns
        -------
        composition: None or Dict
            A dictionary of elements and counts, where count is int unless it is needed
            to be a float.
            Returns None in case of error.

        """
        tmp_formula = self.formula
        if tmp_formula is None:
            return {}
        # necessary for some old pickles which use the deprecated
        # Formula class
        tmp_formula = str(self.formula)
        # commonly occurring characters in incorrectly constructed formulas
        if "*" in tmp_formula:
            warn(f"invalid character '*' found in formula '{self.formula}'")
            tmp_formula = tmp_formula.replace("*", "")
        if "(" in tmp_formula or ")" in tmp_formula:
            warn(f"invalid formula (has parenthesis) in '{self.formula}'")
            return None
        composition = {}
        parsed = element_re.findall(tmp_formula)
        for element, count in parsed:
            if count == "":
                count = 1
            else:
                try:
                    count = float(count)
                    int_count = int(count)
                    if count == int_count:
                        count = int_count
                    else:
                        warn(f"{count} is not an integer (in formula {self.formula})")
                except ValueError:
                    warn(f"failed to parse {count} (in formula {self.formula})")
                    return None
            if element in composition:
                composition[element] += count
            else:
                composition[element] = count
        return composition

    @elements.setter
    def elements(self, elements_dict: Dict[str, Union[int, float]]) -> None:
        """Update formula based on elements dictionary.

        Parameters
        ----------
        elements_dict: dict
            A dicitonary of elements as keys, count as items.
        """

        def stringify(element, number):
            return element if number == 1 else element + str(number)

        self.formula = "".join(
            stringify(e, n) for e, n in sorted(elements_dict.items())
        )

    @property
    def formula_weight(self) -> Union[int, float]:
        """Calculate the formula weight.

        Returns
        ------
        float, int
            Weight of formula, based on the weight and count of elements. Can be int if
            the formula weight is a whole number, but unlikely.
        """
        try:
            return sum(
                [
                    count * elements_and_molecular_weights[element]
                    for element, count in self.elements.items()
                ]
            )
        except KeyError as e:
            warn(f"The element {e} does not appear in the periodic table")

    @property
    def y(self) -> float:
        """Return the shadow price for the metabolite in the most recent solution.

        Shadow prices are computed from the dual values of the bounds in
        the solution.
        .. deprecated ::
        Use metabolite.shadow_price instead.

        Returns
        -------
        float
            Float representing the shadow price.
        """
        warn("Please use metabolite.shadow_price instead.", DeprecationWarning)
        return self.shadow_price

    @property
    def shadow_price(self) -> float:
        """Return the shadow price for the metabolite in the most recent solution.

        Shadow price is the dual value of the corresponding constraint in the
        model.

        Returns
        -------
        shadow_price: float

        Warnings
        --------
        * Accessing shadow prices through a `Solution` object is the safer,
          preferred, and only guaranteed to be correct way. You can see how to
          do so easily in the examples.
        * Shadow price is retrieved from the currently defined
          `self._model.solver`. The solver status is checked but there are no
          guarantees that the current solver state is the one you are looking
          for.
        * If you modify the underlying model after an optimization, you will
          retrieve the old optimization values.

        Raises
        ------
        RuntimeError
            If the underlying model was never optimized beforehand or the
            metabolite is not part of a model.
        OptimizationError
            If the solver status is anything other than 'optimal'.

        Examples
        --------
        >>> from cobra.io import load_model
        >>> model = load_model("textbook")
        >>> solution = model.optimize()
        >>> model.metabolites.glc__D_e.shadow_price
        -0.09166474637510488
        >>> solution.shadow_prices.glc__D_e
        -0.091664746375104883
        """
        try:
            check_solver_status(self._model.solver.status)
            return self._model.constraints[self.id].dual
        except AttributeError:
            raise RuntimeError(f"metabolite '{self.id}' is not part of a model")
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise err.with_traceback()
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise OptimizationError(
                f"Likely no solution exists. Original solver message: {str(err)}."
            ) from err

    def remove_from_model(self, destructive: bool = False) -> None:
        """Remove the association from self.model.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        destructive : bool, default False
            If False then the metabolite is removed from all
            associated reactions.  If True then all associated
            reactions are removed from the Model.
        """
        self._model.remove_metabolites(self, destructive)

    def summary(
        self,
        solution: Optional["Solution"] = None,
        fva: Optional[Union[float, "DataFrame"]] = None,
    ) -> "MetaboliteSummary":
        """Create a summary of the producing and consuming fluxes.

        Parameters
        ----------
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Returns
        -------
        cobra.summary.MetaboliteSummary

        See Also
        --------
        Reaction.summary
        Model.summary

        """
        from cobra.summary import MetaboliteSummary

        return MetaboliteSummary(
            metabolite=self,
            model=self._model,
            solution=solution,
            fva=fva,
        )

    def _repr_html_(self) -> str:
        """Return the metabolite as an HTML string."""
        return f"""
        <table>
            <tr>
                <td><strong>Metabolite identifier</strong></td><td>{self.id}</td>
            </tr><tr>
                <td><strong>Name</strong></td><td>{format_long_string(self.name)}</td>
            </tr><tr>
                <td><strong>Memory address</strong></td>
                <td>{id(self):#x}</td>
            </tr><tr>
                <td><strong>Formula</strong></td><td>{self.formula}</td>
            </tr><tr>
                <td><strong>Compartment</strong></td><td>{self.compartment}</td>
            </tr><tr>
                <td><strong>In {len(self.reactions)} reaction(s)</strong></td><td>
                    {format_long_string(", ".join(r.id for r in self.reactions), 200)}
                    </td>
            </tr>
        </table>"""



"""Provide functions for loading metabolic models from local package data."""


import importlib_resources

import cobra.data

from .abstract_model_repository import AbstractModelRepository


class Cobrapy(AbstractModelRepository):
    """
    Define a concrete implementation of the cobrapy (local package) repository.

     Attributes
    ----------
    name : str
        The name of the Cobrapy repository.
    """

    name: str = "Cobrapy"

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        Initialize a local Cobrapy repository interface.

        Other Parameters
        ----------------
        kwargs
            Passed to the parent constructor in order to enable multiple inheritance.

        """
        super().__init__(url="file://", **kwargs)

    def get_sbml(self, model_id: str) -> bytes:
        """
        Attempt to open an SBML document from the local repository.

        Parameters
        ----------
        model_id : str
            The identifier of the desired metabolic model. This is typically repository
            specific.

        Returns
        -------
        bytes
            A gzip-compressed, UTF-8 encoded SBML document.
        """
        return (
            importlib_resources.files(cobra.data)
            .joinpath(f"{model_id}.xml.gz")
            .read_bytes()
        )



from .annotate import add_SBO
from .delete import (
    delete_model_genes,
    knock_out_model_genes,
    prune_unused_metabolites,
    prune_unused_reactions,
    remove_genes,
)
from .modify import escape_ID, rename_genes
from .validate import check_mass_balance, check_metabolite_compartment_formula
